@@ -116,44 +116,26 @@ static int efa_domain_init_qp_table(struct efa_domain *efa_domain)
 
 static int efa_domain_init_rdm(struct efa_domain *efa_domain, struct fi_info *info)
 {
+	struct fi_info *shm_info = NULL;
 	int err;
-	bool enable_shm = efa_env.enable_shm_transfer;
 
 	assert(EFA_INFO_TYPE_IS_RDM(info));
 
-	/* App provided hints supercede environmental variables.
-	 *
-	 * Using the shm provider comes with some overheads, so avoid
-	 * initializing the provider if the app provides a hint that it does not
-	 * require node-local communication. We can still loopback over the EFA
-	 * device in cases where the app violates the hint and continues
-	 * communicating with node-local peers.
-	 *
-	 */
-	if ((info->caps & FI_REMOTE_COMM)
-	    /* but not local communication */
-	    && !(info->caps & FI_LOCAL_COMM)) {
-		enable_shm = false;
-	}
-
-	efa_domain->shm_info = NULL;
-	if (enable_shm)
-		efa_shm_info_create(info, &efa_domain->shm_info);
-	else
-		EFA_INFO(FI_LOG_CORE, "EFA will not use SHM for intranode communication because FI_EFA_ENABLE_SHM_TRANSFER=0\n");
-
-	if (efa_domain->shm_info) {
-		err = fi_fabric(efa_domain->shm_info->fabric_attr,
+	efa_shm_info_create(info, &shm_info);
+	if (shm_info && !efa_domain->fabric->shm_fabric) {
+		err = fi_fabric(shm_info->fabric_attr,
 				&efa_domain->fabric->shm_fabric,
 				efa_domain->fabric->util_fabric.fabric_fid.fid.context);
-		if (err)
+		if (err) {
+			EFA_WARN(FI_LOG_DOMAIN, 
+				 "Failed to create shm_fabric: %s\n",
+				 fi_strerror(-err));
 			return err;
-	} else {
-		efa_domain->fabric->shm_fabric = NULL;
+		}
 	}
 
 	if (efa_domain->fabric->shm_fabric) {
-		err = fi_domain(efa_domain->fabric->shm_fabric, efa_domain->shm_info,
+		err = fi_domain(efa_domain->fabric->shm_fabric, shm_info,
 				&efa_domain->shm_domain, NULL);
 		if (err)
 			return err;
@@ -169,6 +151,10 @@ static int efa_domain_init_rdm(struct efa_domain *efa_domain, struct fi_info *in
 	dlist_init(&efa_domain->ope_longcts_send_list);
 	dlist_init(&efa_domain->peer_backoff_list);
 	dlist_init(&efa_domain->handshake_queued_peer_list);
+
+	if (shm_info)
+		fi_freeinfo(shm_info);
+
 	return 0;
 }
 
@@ -390,9 +376,6 @@ static int efa_domain_close(fid_t fid)
 			return ret;
 	}
 
-	if (efa_domain->shm_info)
-		fi_freeinfo(efa_domain->shm_info);
-
 	if (efa_domain->info)
 		fi_freeinfo(efa_domain->info);
 
@@ -606,6 +589,12 @@ static void efa_cq_ext_progress_no_op(struct util_cq *cq)
 	return;
 }
 
+// Avoid flushing cq when it is created with external memory
+static int efa_cq_ext_no_poll_ibv_cq(ssize_t cqe_to_process, struct efa_ibv_cq *ibv_cq)
+{
+	return ENOENT;
+}
+
 /**
  * @brief Create a completion queue with external memory provided via dmabuf.
  *
@@ -625,6 +614,8 @@ static int efa_domain_cq_open_ext(struct fid_domain *domain_fid,
 	struct efa_domain *efa_domain;
 	int err, retv;
 
+	/* GPU cannot do a blocking wait on CQ entries because 
+	 * system FDs are only accessible to CPU. */
 	if (attr->wait_obj != FI_WAIT_NONE)
 		return -FI_ENOSYS;
 
@@ -651,6 +642,7 @@ static int efa_domain_cq_open_ext(struct fid_domain *domain_fid,
 	 * CQ polling is safe when CPU virtual address is provided in buffer.
 	 * Otherwise, the memory is on GPU and the use of CQ poll interfaces should be avoided.
 	 */
+	cq->poll_ibv_cq = efa_cq_init_attr->ext_mem_dmabuf.buffer ? efa_cq_poll_ibv_cq: efa_cq_ext_no_poll_ibv_cq;
 	err = ofi_cq_init(&efa_prov, domain_fid, attr, &cq->util_cq,
 			  efa_cq_init_attr->ext_mem_dmabuf.buffer ?
 				  &efa_cq_progress :
@@ -663,15 +655,16 @@ static int efa_domain_cq_open_ext(struct fid_domain *domain_fid,
 
 	efa_domain = container_of(cq->util_cq.domain, struct efa_domain,
 				  util_domain);
-	err = efa_cq_ibv_cq_ex_open(attr, efa_domain->device->ibv_ctx,
-				    &cq->ibv_cq.ibv_cq_ex,
-				    &cq->ibv_cq.ibv_cq_ex_type,
+	err = efa_cq_open_ibv_cq(attr, efa_domain->device->ibv_ctx,
+				    &cq->ibv_cq,
 				    efa_cq_init_attr);
 	if (err) {
 		EFA_WARN(FI_LOG_CQ, "Unable to create extended CQ with external memory: %s\n", fi_strerror(err));
 		goto err_free_util_cq;
 	}
 
+	ofi_atomic_initialize32(&cq->nevents, 0);
+
 	*cq_fid = &cq->util_cq.cq_fid;
 	(*cq_fid)->fid.fclass = FI_CLASS_CQ;
 	(*cq_fid)->fid.context = context;
@@ -699,12 +692,29 @@ static int efa_domain_cq_open_ext(struct fid_domain *domain_fid,
 }
 #endif
 
+static uint64_t efa_domain_get_mr_lkey(struct fid_mr *mr)
+{
+	struct efa_mr *efa_mr;
+
+	efa_mr = container_of(mr, struct efa_mr, mr_fid);
+	if (!efa_mr->ibv_mr) {
+		EFA_WARN(FI_LOG_DOMAIN, "MR not registered\n");
+		return FI_KEY_NOTAVAIL;
+	}
+
+	return efa_mr->ibv_mr->lkey;
+}
+
 static struct fi_efa_ops_domain efa_ops_domain = {
 	.query_mr = efa_domain_query_mr,
+};
+
+static struct fi_efa_ops_gda efa_ops_gda = {
 	.query_addr = efa_domain_query_addr,
 	.query_qp_wqs = efa_domain_query_qp_wqs,
 	.query_cq = efa_domain_query_cq,
 	.cq_open_ext = efa_domain_cq_open_ext,
+	.get_mr_lkey = efa_domain_get_mr_lkey,
 };
 
 static int
@@ -712,14 +722,25 @@ efa_domain_ops_open(struct fid *fid, const char *ops_name, uint64_t flags,
 		     void **ops, void *context)
 {
 	int ret = FI_SUCCESS;
+	struct efa_domain *efa_domain;
 
 	if (strcmp(ops_name, FI_EFA_DOMAIN_OPS) == 0) {
 		*ops = &efa_ops_domain;
-	} else {
-		EFA_WARN(FI_LOG_DOMAIN,
-			"Unknown ops name: %s\n", ops_name);
-		ret = -FI_EINVAL;
+		return ret;
 	}
+	if (strcmp(ops_name, FI_EFA_GDA_OPS) == 0) {
+		efa_domain = container_of(fid, struct efa_domain, util_domain.domain_fid.fid);
+		if (efa_domain->info_type != EFA_INFO_DIRECT) {
+			EFA_WARN(FI_LOG_DOMAIN, "Only efa direct supports FI_EFA_GDA_OPS\n");
+			return -FI_EOPNOTSUPP;
+		}
+
+		*ops = &efa_ops_gda;
+		return ret;
+	}
+
+	EFA_WARN(FI_LOG_DOMAIN, "Unknown ops name: %s\n", ops_name);
+	ret = -FI_EINVAL;
 
 	return ret;
 }
@@ -759,8 +780,11 @@ void efa_domain_progress_rdm_peers_and_queues(struct efa_domain *domain)
 
 		if (OFI_UNLIKELY(ret)) {
 			EFA_WARN(FI_LOG_EP_CTRL,
-				"Failed to post HANDSHAKE to peer %ld: %s\n",
-				peer->efa_fiaddr, fi_strerror(-ret));
+				 "Failed to post HANDSHAKE to peer fi_addr: "
+				 "%ld implicit fi_addr: %ld. %s\n",
+				 peer->conn->fi_addr,
+				 peer->conn->implicit_fi_addr,
+				 fi_strerror(-ret));
 			efa_base_ep_write_eq_error(&peer->ep->base_ep, -ret, FI_EFA_ERR_PEER_HANDSHAKE);
 			continue;
 		}
@@ -776,100 +800,26 @@ void efa_domain_progress_rdm_peers_and_queues(struct efa_domain *domain)
 	dlist_foreach_container_safe(&domain->ope_queued_list,
 				     struct efa_rdm_ope,
 				     ope, queued_entry, tmp) {
-		peer = efa_rdm_ep_get_peer(ope->ep, ope->addr);
 
+		peer = ope->peer;
 		if (peer && (peer->flags & EFA_RDM_PEER_IN_BACKOFF))
 			continue;
 
-		if (ope->internal_flags & EFA_RDM_OPE_QUEUED_BEFORE_HANDSHAKE) {
-			ret = efa_rdm_ope_repost_ope_queued_before_handshake(ope);
-			if (ret == -FI_EAGAIN)
-				continue;
-
-			if (OFI_UNLIKELY(ret)) {
-				assert(ope->type == EFA_RDM_TXE);
-				/* efa_rdm_txe_handle_error will remove ope from the queued_list */
-				ope->ep->ope_queued_before_handshake_cnt--;
-				efa_rdm_txe_handle_error(ope, -ret, FI_EFA_ERR_PKT_POST);
-				continue;
-			}
-
-			dlist_remove(&ope->queued_entry);
-			ope->internal_flags &= ~EFA_RDM_OPE_QUEUED_BEFORE_HANDSHAKE;
-			ope->ep->ope_queued_before_handshake_cnt--;
-		}
-
-		if (ope->internal_flags & EFA_RDM_OPE_QUEUED_RNR) {
-			assert(!dlist_empty(&ope->queued_pkts));
-			ret = efa_rdm_ep_post_queued_pkts(ope->ep, &ope->queued_pkts);
-
-			if (ret == -FI_EAGAIN)
-				continue;
-
-			if (OFI_UNLIKELY(ret)) {
-				assert(ope->type == EFA_RDM_RXE || ope->type == EFA_RDM_TXE);
-				if (ope->type == EFA_RDM_RXE)
-					efa_rdm_rxe_handle_error(ope, -ret, FI_EFA_ERR_PKT_SEND);
-				else
-					efa_rdm_txe_handle_error(ope, -ret, FI_EFA_ERR_PKT_SEND);
-				continue;
-			}
-
-			dlist_remove(&ope->queued_entry);
-			ope->internal_flags &= ~EFA_RDM_OPE_QUEUED_RNR;
-		}
-
-		if (ope->internal_flags & EFA_RDM_OPE_QUEUED_CTRL) {
-			ret = efa_rdm_ope_post_send(ope, ope->queued_ctrl_type);
-			if (ret == -FI_EAGAIN)
-				continue;
-
-			if (OFI_UNLIKELY(ret)) {
-				assert(ope->type == EFA_RDM_TXE || ope->type == EFA_RDM_RXE);
-				if (ope->type == EFA_RDM_TXE)
-					efa_rdm_txe_handle_error(ope, -ret, FI_EFA_ERR_PKT_POST);
-				else
-					efa_rdm_rxe_handle_error(ope, -ret, FI_EFA_ERR_PKT_POST);
-				continue;
-			}
-
-			/* it can happen that efa_rdm_ope_post_send() released ope
-			 * (if the ope is rxe and packet type is EOR and inject is used). In
-			 * that case rxe's state has been set to EFA_RDM_OPE_FREE and
-			 * it has been removed from ep->op_queued_entry_list, so nothing
-			 * is left to do.
-			 */
-			if (ope->state == EFA_RDM_OPE_FREE)
-				continue;
-
-			ope->internal_flags &= ~EFA_RDM_OPE_QUEUED_CTRL;
-			dlist_remove(&ope->queued_entry);
-		}
-
-		if (ope->internal_flags & EFA_RDM_OPE_QUEUED_READ) {
-			ret = efa_rdm_ope_post_read(ope);
-			if (ret == -FI_EAGAIN)
-				continue;
-
-			if (OFI_UNLIKELY(ret)) {
-				assert(ope->type == EFA_RDM_TXE || ope->type == EFA_RDM_RXE);
-				if (ope->type == EFA_RDM_TXE)
-					efa_rdm_txe_handle_error(ope, -ret, FI_EFA_ERR_READ_POST);
-				else
-					efa_rdm_rxe_handle_error(ope, -ret, FI_EFA_ERR_READ_POST);
-				continue;
-			}
-
-			ope->internal_flags &= ~EFA_RDM_OPE_QUEUED_READ;
-			dlist_remove(&ope->queued_entry);
-		}
+		if (efa_rdm_ope_process_queued_ope(ope, EFA_RDM_OPE_QUEUED_BEFORE_HANDSHAKE))
+			continue;
+		if (efa_rdm_ope_process_queued_ope(ope, EFA_RDM_OPE_QUEUED_RNR))
+			continue;
+		if (efa_rdm_ope_process_queued_ope(ope, EFA_RDM_OPE_QUEUED_CTRL))
+			continue;
+		if (efa_rdm_ope_process_queued_ope(ope, EFA_RDM_OPE_QUEUED_READ))
+			continue;
 	}
 	/*
 	 * Send data packets until window or data queue is exhausted.
 	 */
 	dlist_foreach_container(&domain->ope_longcts_send_list, struct efa_rdm_ope,
 				ope, entry) {
-		peer = efa_rdm_ep_get_peer(ope->ep, ope->addr);
+		peer = ope->peer;
 		assert(peer);
 		if (peer->flags & EFA_RDM_PEER_IN_BACKOFF)
 			continue;
