@@ -3,6 +3,7 @@
 
 #include "efa.h"
 #include "efa_rdm_cq.h"
+#include "efa_data_path_ops.h"
 #include "ofi_util.h"
 #include "efa_av.h"
 #include "efa_cntr.h"
@@ -87,7 +88,7 @@ static struct fi_ops efa_rdm_cq_fi_ops = {
  */
 static
 void efa_rdm_cq_proc_ibv_recv_rdma_with_imm_completion(
-						       struct ibv_cq_ex *ibv_cq_ex,
+						       struct efa_ibv_cq *ibv_cq,
 						       uint64_t flags,
 						       struct efa_rdm_ep *ep,
 						       struct efa_rdm_pke *pkt_entry
@@ -97,16 +98,18 @@ void efa_rdm_cq_proc_ibv_recv_rdma_with_imm_completion(
 	int ret;
 	fi_addr_t src_addr;
 	struct efa_av *efa_av;
-	uint32_t imm_data = ibv_wc_read_imm_data(ibv_cq_ex);
-	uint32_t len = ibv_wc_read_byte_len(ibv_cq_ex);
+	uint32_t imm_data = efa_ibv_cq_wc_read_imm_data(ibv_cq);
+	uint32_t len = efa_ibv_cq_wc_read_byte_len(ibv_cq);
 
 	target_cq = ep->base_ep.util_ep.rx_cq;
 	efa_av = ep->base_ep.av;
 
 	if (ep->base_ep.util_ep.caps & FI_SOURCE) {
+
+		/* Only check the explicit AV when writing completions */
 		src_addr = efa_av_reverse_lookup_rdm(efa_av,
-						ibv_wc_read_slid(ibv_cq_ex),
-						ibv_wc_read_src_qp(ibv_cq_ex),
+						efa_ibv_cq_wc_read_slid(ibv_cq),
+						efa_ibv_cq_wc_read_src_qp(ibv_cq),
 						NULL);
 		ret = ofi_cq_write_src(target_cq, NULL, flags, len, NULL, imm_data, 0, src_addr);
 	} else {
@@ -126,7 +129,7 @@ void efa_rdm_cq_proc_ibv_recv_rdma_with_imm_completion(
 	 * For unsolicited wc, pkt_entry can be NULL, so we can only
 	 * access it for solicited wc.
 	 */
-	if (!efa_cq_wc_is_unsolicited(ibv_cq_ex)) {
+	if (!efa_cq_wc_is_unsolicited(ibv_cq)) {
 		/**
 		 * Recv with immediate will consume a pkt_entry, but the pkt is not
 		 * filled, so free the pkt_entry and record we have one less posted
@@ -138,112 +141,279 @@ void efa_rdm_cq_proc_ibv_recv_rdma_with_imm_completion(
 	}
 }
 
-#if HAVE_EFADV_CQ_EX
 /**
- * @brief Read peer raw address from EFA device and look up the peer address in AV.
- * This function should only be called if the peer AH is unknown.
- * @return Peer address, or FI_ADDR_NOTAVAIL if unavailable.
+ * @brief Populate the efa_ep_addr struct with the raw address, QPN and QKEY
+ * This function tries to read the raw address first from the packet header
+ * If the packet header does not have the raw address, it uses
+ * efadv_wc_read_sgid if available.
+ * @return FI_SUCCESS if the efa_ep_addr struct is populated. Error otherwise.
  */
-static inline
-fi_addr_t efa_rdm_cq_determine_peer_address_from_efadv(
-						       struct ibv_cq_ex *ibv_cqx,
-						       enum ibv_cq_ex_type ibv_cq_ex_type)
+static inline int efa_rdm_cq_populate_src_efa_ep_addr(
+	struct efa_rdm_pke *pkt_entry, struct efa_ibv_cq *ibv_cq,
+	struct efa_ep_addr *efa_ep_addr)
 {
-	struct efa_rdm_pke *pkt_entry;
+	struct efa_rdm_base_hdr *base_hdr;
 	struct efa_rdm_ep *ep;
-	struct efa_ep_addr efa_ep_addr = {0};
-	fi_addr_t addr;
-	union ibv_gid gid = {0};
-	uint32_t *connid = NULL;
+	void *raw_addr_ptr = NULL;
+	uint32_t *connid;
 
-	if (ibv_cq_ex_type != EFADV_CQ) {
-		/* EFA DV CQ is not supported. This could be due to old EFA kernel module versions. */
-		return FI_ADDR_NOTAVAIL;
-	}
-
-	/* Attempt to read sgid from EFA firmware */
-	if (efadv_wc_read_sgid(efadv_cq_from_ibv_cq_ex(ibv_cqx), &gid) < 0) {
-		/* Return code is negative if the peer AH is known */
-		return FI_ADDR_NOTAVAIL;
-	}
-
-	pkt_entry = (void *)(uintptr_t)ibv_cqx->wr_id;
 	ep = pkt_entry->ep;
 	assert(ep);
 
+	efa_ep_addr->qpn = efa_ibv_cq_wc_read_src_qp(ibv_cq);
+
+	if (pkt_entry->alloc_type == EFA_RDM_PKE_FROM_USER_RX_POOL) {
+		/* Receive packet posted in the zero-copy path does not have a
+		 * header. So we cannot read the connid or the raw address. */
+		return FI_EADDRNOTAVAIL;
+	}
+
 	connid = efa_rdm_pke_connid_ptr(pkt_entry);
 	if (!connid) {
-		return FI_ADDR_NOTAVAIL;
+		return FI_EADDRNOTAVAIL;
+	}
+	efa_ep_addr->qkey = *connid;
+
+	base_hdr = efa_rdm_pke_get_base_hdr(pkt_entry);
+	if (base_hdr->version < EFA_RDM_PROTOCOL_VERSION) {
+		char self_raw_addr_str[OFI_ADDRSTRLEN];
+		size_t buflen = OFI_ADDRSTRLEN;
+
+		efa_base_ep_raw_addr_str(&ep->base_ep, self_raw_addr_str,
+					 &buflen);
+		EFA_WARN(FI_LOG_CQ,
+			 "Host %s received a packet with invalid protocol "
+			 "version %d.\n"
+			 "This host can only support protocol version %d and "
+			 "above.\n",
+			 self_raw_addr_str, base_hdr->version,
+			 EFA_RDM_PROTOCOL_VERSION);
+		efa_base_ep_write_eq_error(&ep->base_ep, FI_EIO,
+					   FI_EFA_ERR_INVALID_PKT_TYPE);
+		fprintf(stderr,
+			"Host %s received a packet with invalid protocol "
+			"version %d.\n"
+			"This host can only support protocol version %d and "
+			"above. %s:%d\n",
+			self_raw_addr_str, base_hdr->version,
+			EFA_RDM_PROTOCOL_VERSION, __FILE__, __LINE__);
+		abort();
 	}
 
-	/*
-	 * Use raw:qpn:connid as the key to lookup AV for peer's fi_addr
-	 */
-	memcpy(efa_ep_addr.raw, gid.raw, sizeof(efa_ep_addr.raw));
-	efa_ep_addr.qpn = ibv_wc_read_src_qp(ibv_cqx);
-	efa_ep_addr.qkey = *connid;
-	addr = ofi_av_lookup_fi_addr(&ep->base_ep.av->util_av, &efa_ep_addr);
-	if (addr != FI_ADDR_NOTAVAIL) {
-		char gid_str_cdesc[INET6_ADDRSTRLEN];
-		inet_ntop(AF_INET6, gid.raw, gid_str_cdesc, INET6_ADDRSTRLEN);
-		EFA_WARN(FI_LOG_AV,
-				"Recovered peer fi_addr. [Raw]:[QPN]:[QKey] = [%s]:[%" PRIu16 "]:[%" PRIu32 "]\n",
-				gid_str_cdesc, efa_ep_addr.qpn, efa_ep_addr.qkey);
+	/* Attempt to read raw address from packet header */
+	if (base_hdr->type >= EFA_RDM_REQ_PKT_BEGIN &&
+	    efa_rdm_pke_get_req_raw_addr(pkt_entry)) {
+		raw_addr_ptr = efa_rdm_pke_get_req_raw_addr(pkt_entry);
+		assert(raw_addr_ptr);
+		goto out;
 	}
 
-	return addr;
-}
-
-/**
- * @brief Determine peer address from ibv_cq_ex
- * Attempt to inject or determine peer address if not available. This usually
- * happens when the endpoint receives the first packet from a new peer.
- * There is an edge case for EFA endpoint - the device might lose the address
- * handle of a known peer due to a firmware bug and return FI_ADDR_NOTAVAIL.
- * The provider needs to look up the address using Raw address:QPN:QKey.
- * Note: This function introduces addtional overhead. It should only be called if
- * efa_av_lookup_address_rdm fails to find the peer address.
- * @param ep Pointer to RDM endpoint
- * @param ibv_cqx Pointer to CQ
- * @returns Peer address, or FI_ADDR_NOTAVAIL if unsuccessful.
- */
-static inline fi_addr_t efa_rdm_cq_determine_addr_from_ibv_cq(struct ibv_cq_ex *ibv_cqx, enum ibv_cq_ex_type ibv_cq_ex_type)
-{
-	struct efa_rdm_pke *pkt_entry;
-	fi_addr_t addr = FI_ADDR_NOTAVAIL;
-
-	pkt_entry = (void *)(uintptr_t)ibv_cqx->wr_id;
-
-	addr = efa_rdm_pke_determine_addr(pkt_entry);
-
-	if (addr == FI_ADDR_NOTAVAIL) {
-		addr = efa_rdm_cq_determine_peer_address_from_efadv(ibv_cqx, ibv_cq_ex_type);
+#if HAVE_EFADV_CQ_EX
+	union ibv_gid gid = {0};
+	if (ibv_cq->ibv_cq_ex_type != EFADV_CQ) {
+		/* EFA DV CQ is not supported. This could be due to old EFA
+		 * kernel module versions. */
+		return FI_EADDRNOTAVAIL;
 	}
 
-	return addr;
-}
-#else
-/**
- * @brief Determine peer address from ibv_cq_ex
- * Attempt to inject peer address if not available. This usually
- * happens when the endpoint receives the first packet from a new peer.
- * Note: This function introduces addtional overhead. It should only be called if
- * efa_av_lookup_address_rdm fails to find the peer address.
- * @param ep Pointer to RDM endpoint
- * @param ibv_cqx Pointer to CQ
- * @returns Peer address, or FI_ADDR_NOTAVAIL if unsuccessful.
- */
-static inline
-fi_addr_t efa_rdm_cq_determine_addr_from_ibv_cq(struct ibv_cq_ex *ibv_cqx, enum ibv_cq_ex_type ibv_cq_ex_type)
-{
-	struct efa_rdm_pke *pkt_entry;
+	/* Attempt to read sgid from EFA firmware */
+	if (efa_ibv_cq_wc_read_sgid(ibv_cq, &gid) < 0) {
+		/* Return code is negative if the peer AH is known */
+		return FI_EADDRNOTAVAIL;
+	}
 
-	pkt_entry = (void *)(uintptr_t)ibv_cqx->wr_id;
-
-	return efa_rdm_pke_determine_addr(pkt_entry);
-}
+	raw_addr_ptr = &gid.raw;
 #endif
 
+out:
+	if (raw_addr_ptr) {
+		memcpy(efa_ep_addr->raw, raw_addr_ptr,
+		       sizeof(efa_ep_addr->raw));
+		return FI_SUCCESS;
+	}
+
+	return FI_EADDRNOTAVAIL;
+}
+
+/**
+ * @brief Read peer raw address from packet entry or the EFA device and look up
+ * the peer address in explicit and implicit AVs This function should only be
+ * called if the peer AH is unknown.
+ * @return Pointer to peer struct, or NULL if unavailable.
+ */
+static inline struct efa_rdm_peer *
+efa_rdm_cq_lookup_raw_addr(struct efa_rdm_pke *pke,
+			   struct efa_ep_addr *efa_ep_addr)
+{
+	struct efa_rdm_ep *ep;
+	fi_addr_t addr;
+	struct efa_rdm_peer *peer = NULL;
+	bool implicit = false;
+	char gid_str_cdesc[INET6_ADDRSTRLEN];
+
+	ep = pke->ep;
+	assert(ep);
+
+	/* First check the explicit AV */
+	addr = ofi_av_lookup_fi_addr(&ep->base_ep.av->util_av,
+				     (void *) efa_ep_addr);
+	if (addr != FI_ADDR_NOTAVAIL) {
+		implicit = false;
+		peer = efa_rdm_ep_get_peer(ep, addr);
+		assert(peer);
+		goto out;
+	}
+
+	/* Next check implicit AV */
+	addr = ofi_av_lookup_fi_addr(&ep->base_ep.av->util_av_implicit,
+				     (void *) efa_ep_addr);
+	if (addr != FI_ADDR_NOTAVAIL) {
+		implicit = true;
+		peer = efa_rdm_ep_get_peer_implicit(ep, addr);
+		assert(peer);
+		goto out;
+	}
+
+	return NULL;
+
+out:
+	inet_ntop(AF_INET6, efa_ep_addr->raw, gid_str_cdesc, INET6_ADDRSTRLEN);
+	EFA_WARN(FI_LOG_AV,
+		 "Recovered fi_addr for peer:[QPN]:[QKey] = "
+		 "[%s]:[%" PRIu16 "]:[%" PRIu32 "] fi_addr: %" PRIu64
+		 " implicit AV: %s\n",
+		 gid_str_cdesc, efa_ep_addr->qpn, efa_ep_addr->qkey, addr,
+		 implicit ? "true" : "false");
+
+	return peer;
+}
+
+/**
+ * @brief Determine peer struct from ibv_cq and packet entry
+ * Insert the peer into the implicit AV if not found. This
+ * happens when the endpoint receives the first packet from a new peer.
+ * @param ep Pointer to RDM endpoint
+ * @param ibv_cq Pointer to CQ
+ * @param pkt_entry packet entry
+ * @returns Peer address, or FI_ADDR_NOTAVAIL if unsuccessful.
+ */
+static inline struct efa_rdm_peer *
+efa_rdm_cq_get_peer_for_pkt_entry(struct efa_rdm_ep *ep,
+				  struct efa_ibv_cq *efa_ibv_cq,
+				  struct efa_rdm_pke *pkt_entry)
+{
+	struct efa_av *efa_av = ep->base_ep.av;
+	fi_addr_t explicit_fi_addr, implicit_fi_addr;
+	struct efa_ep_addr efa_ep_addr = {0};
+	struct efa_ep_addr_hashable *efa_ep_addr_hashable = NULL;
+	struct efa_rdm_peer *peer = NULL;
+	int ret;
+	uint32_t gid;
+	uint32_t qpn;
+
+	gid = efa_ibv_cq_wc_read_slid(efa_ibv_cq);
+	qpn = efa_ibv_cq_wc_read_src_qp(efa_ibv_cq);
+
+	EFA_DBG(FI_LOG_CQ,
+		"Attempting to retrieve peer for packet from peer with gid %d "
+		"and qpn %d\n",
+		gid, qpn);
+
+	/* To determine the source peer struct, the workflow is the following
+	 * 1. Get GID and QPN from rdma-core and check the explicit AV
+	 * 2. If not found, check the implicit AV with GID and QPN
+	 * 3 (a). If not found, retrieve raw address from the packet header
+	 * 3 (b). If packet header doesn't have raw address, retrieve raw
+	 * address with efadv_wc_read_sgid
+	 * 4. Check explicit and implicit AVs for the raw address retrieved in
+	 * (3)
+	 * 5. If not found and raw address is available, insert raw address into
+	 * the implicit AV
+	 *
+	 * TODO: Remove the usage of efadv_wc_read_sgid after EFA device's
+	 * behavior is fixed
+	 */
+	explicit_fi_addr =
+		efa_av_reverse_lookup_rdm(efa_av, gid, qpn, pkt_entry);
+
+	if (explicit_fi_addr != FI_ADDR_NOTAVAIL) {
+		EFA_DBG(FI_LOG_CQ,
+			"Peer with gid %d and qpn %d found in explicit AV with "
+			"fi_addr %ld\n",
+			gid, qpn, explicit_fi_addr);
+		peer = efa_rdm_ep_get_peer(ep, explicit_fi_addr);
+		goto out;
+	}
+
+	implicit_fi_addr =
+		efa_av_reverse_lookup_rdm_implicit(efa_av, gid, qpn, pkt_entry);
+
+	if (implicit_fi_addr != FI_ADDR_NOTAVAIL) {
+		EFA_DBG(FI_LOG_CQ,
+			"Peer with gid %d and qpn %d found in implicit AV with "
+			"fi_addr %ld\n",
+			gid, qpn, implicit_fi_addr);
+		peer = efa_rdm_ep_get_peer_implicit(ep, implicit_fi_addr);
+		goto out;
+	}
+
+	ret = efa_rdm_cq_populate_src_efa_ep_addr(
+		pkt_entry, efa_ibv_cq,
+		&efa_ep_addr);
+	if (ret) {
+		/* Failed to read raw address from packet entry and
+		 * efadv_wc_read_sgid */
+		return NULL;
+	}
+
+	/* If the packet is from a peer that we evicted from the implicit AV,
+	 * print a warning and ignore the packet. We do this because we lose
+	 * information about previous communication from the peer when we evict
+	 * the peer from the implicit AV
+	 *
+	 * TODO: continue communication with peer by saving the previous state
+	 * and restoring it
+	 */
+	HASH_FIND(hh, ep->base_ep.av->evicted_peers_hashset, &efa_ep_addr,
+		  sizeof(struct efa_ep_addr), efa_ep_addr_hashable);
+	if (OFI_UNLIKELY(!!efa_ep_addr_hashable)) {
+		EFA_WARN(FI_LOG_CQ, "Received packet from peer already evicted "
+				    "from the implicit AV\n");
+		return NULL;
+	}
+
+	peer = efa_rdm_cq_lookup_raw_addr(pkt_entry, &efa_ep_addr);
+	if (peer)
+		goto out;
+
+	EFA_DBG(FI_LOG_CQ,
+		"Peer with gid %d and qpn %d not found in explicit or implicit "
+		"AV. Attempting to insert into implicit AV...\n",
+		gid, qpn);
+	/*
+	 * The message is from a peer through efa device, which means peer is
+	 * not local or shm is disabled for transmission. We shouldn't insert
+	 * in to shm av in this case.
+	 */
+	ret = efa_av_insert_one(ep->base_ep.av, &efa_ep_addr, &implicit_fi_addr,
+				0, NULL, false, true);
+	if (OFI_UNLIKELY(ret != 0)) {
+		efa_base_ep_write_eq_error(&ep->base_ep, ret,
+					   FI_EFA_ERR_AV_INSERT);
+		return NULL;
+	}
+	assert(implicit_fi_addr != FI_ADDR_NOTAVAIL);
+	peer = efa_rdm_ep_get_peer_implicit(ep, implicit_fi_addr);
+
+out:
+	assert(peer);
+	assert((peer->conn->fi_addr != FI_ADDR_NOTAVAIL &&
+		peer->conn->implicit_fi_addr == FI_ADDR_NOTAVAIL) ||
+	       (peer->conn->implicit_fi_addr != FI_ADDR_NOTAVAIL &&
+		peer->conn->fi_addr == FI_ADDR_NOTAVAIL));
+	return peer;
+}
+
 /**
  * @brief handle a received packet
  *
@@ -253,12 +423,9 @@ fi_addr_t efa_rdm_cq_determine_addr_from_ibv_cq(struct ibv_cq_ex *ibv_cqx, enum
 static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct efa_rdm_pke *pkt_entry, struct efa_rdm_ep *ep)
 {
 	int pkt_type;
-	struct efa_rdm_peer *peer;
 	struct efa_rdm_base_hdr *base_hdr;
-	struct efa_av *efa_av = ep->base_ep.av;
 	uint32_t imm_data = 0;
 	bool has_imm_data = false;
-	struct ibv_cq_ex *ibv_cq_ex = ibv_cq->ibv_cq_ex;
 
 	if (pkt_entry->alloc_type == EFA_RDM_PKE_FROM_USER_RX_POOL) {
 		assert(ep->user_rx_pkts_posted > 0);
@@ -268,26 +435,21 @@ static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct
 		ep->efa_rx_pkts_posted--;
 	}
 
-	pkt_entry->addr = efa_av_reverse_lookup_rdm(efa_av, ibv_wc_read_slid(ibv_cq_ex),
-					ibv_wc_read_src_qp(ibv_cq_ex), pkt_entry);
-
-	if (pkt_entry->addr == FI_ADDR_NOTAVAIL) {
-		pkt_entry->addr = efa_rdm_cq_determine_addr_from_ibv_cq(ibv_cq_ex, ibv_cq->ibv_cq_ex_type);
-	}
-
-	pkt_entry->pkt_size = ibv_wc_read_byte_len(ibv_cq_ex);
-	if (ibv_wc_read_wc_flags(ibv_cq_ex) & IBV_WC_WITH_IMM) {
+	pkt_entry->pkt_size = efa_ibv_cq_wc_read_byte_len(ibv_cq);
+	if (efa_ibv_cq_wc_read_wc_flags(ibv_cq) & IBV_WC_WITH_IMM) {
 		has_imm_data = true;
-		imm_data = ibv_wc_read_imm_data(ibv_cq_ex);
+		imm_data = efa_ibv_cq_wc_read_imm_data(ibv_cq);
 	}
 
+	pkt_entry->peer = efa_rdm_cq_get_peer_for_pkt_entry(ep, ibv_cq, pkt_entry);
+
 	/*
 	 * Ignore packet if peer address cannot be determined. This ususally happens if
 	 * we had prior communication with the peer, but
 	 * application called fi_av_remove() to remove the address
 	 * from address vector.
 	 */
-	if (pkt_entry->addr == FI_ADDR_NOTAVAIL) {
+	if (!pkt_entry->peer) {
 		EFA_WARN(FI_LOG_CQ,
 			"Warning: ignoring a received packet from a removed address. packet type: %" PRIu8
 			", packet flags: %x\n",
@@ -304,19 +466,17 @@ static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct
 	efa_rdm_pke_print(pkt_entry, "Received");
 #endif
 #endif
-	peer = efa_rdm_ep_get_peer(ep, pkt_entry->addr);
-	assert(peer);
-	if (peer->is_local) {
+	if (pkt_entry->peer->is_local) {
 		/*
 		 * This happens when the peer is on same instance, but chose to
 		 * use EFA device to communicate with me. In this case, we respect
 		 * that and will not use shm with the peer.
 		 * TODO: decide whether to use shm through handshake packet.
 		 */
-		peer->is_local = 0;
+		pkt_entry->peer->is_local = 0;
 	}
 
-	efa_rdm_ep_post_handshake_or_queue(ep, peer);
+	efa_rdm_ep_post_handshake_or_queue(ep, pkt_entry->peer);
 
 	/**
 	 * Data is already delivered to user posted pkt without pkt hdrs.
@@ -324,7 +484,7 @@ static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct
 	if (pkt_entry->alloc_type == EFA_RDM_PKE_FROM_USER_RX_POOL) {
 		assert(ep->base_ep.user_recv_qp);
 		/* User recv pkts are only posted to the user recv qp */
-		assert(ibv_wc_read_qp_num(ibv_cq->ibv_cq_ex) == ep->base_ep.user_recv_qp->qp_num);
+		assert(efa_ibv_cq_wc_read_qp_num(ibv_cq) == ep->base_ep.user_recv_qp->qp_num);
 		return efa_rdm_pke_proc_received_no_hdr(pkt_entry, has_imm_data, imm_data);
 	}
 
@@ -333,8 +493,11 @@ static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct
 	pkt_type = base_hdr->type;
 	if (OFI_UNLIKELY(pkt_type >= EFA_RDM_EXTRA_REQ_PKT_END)) {
 		EFA_WARN(FI_LOG_CQ,
-			"Peer %d is requesting feature %d, which this EP does not support.\n",
-			(int)pkt_entry->addr, base_hdr->type);
+			 "Peer fi_addr: %ld implicit fi_addr %ld is requesting "
+			 "feature %d, which this EP does not support.\n",
+			 pkt_entry->peer->conn->fi_addr,
+			 pkt_entry->peer->conn->implicit_fi_addr,
+			 base_hdr->type);
 
 		assert(0 && "invalid REQ packet type");
 		efa_base_ep_write_eq_error(&ep->base_ep, FI_EIO, FI_EFA_ERR_INVALID_PKT_TYPE);
@@ -353,21 +516,21 @@ static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct
 		size_t errbuf_len;
 
 		/* local & peer host-id & ep address will be logged by efa_rdm_write_error_msg */
-		if (!efa_rdm_write_error_msg(ep, pkt_entry->addr, FI_EFA_ERR_INVALID_PKT_TYPE_ZCPY_RX, errbuf, &errbuf_len))
+		if (!efa_rdm_write_error_msg(ep, pkt_entry->peer, FI_EFA_ERR_INVALID_PKT_TYPE_ZCPY_RX, errbuf, &errbuf_len))
 			EFA_WARN(FI_LOG_CQ, "Error: %s\n", (const char *) errbuf);
 		efa_base_ep_write_eq_error(&ep->base_ep, FI_EINVAL, FI_EFA_ERR_INVALID_PKT_TYPE_ZCPY_RX);
 		efa_rdm_pke_release_rx(pkt_entry);
 		return;
 	}
 
-	efa_rdm_pke_proc_received(pkt_entry, peer);
+	efa_rdm_pke_proc_received(pkt_entry);
 }
 
 
 /**
  * @brief Get the vendor error code for an endpoint's CQ
  *
- * This function is essentially a wrapper for `ibv_wc_read_vendor_err()`; making
+ * This function is essentially a wrapper for `ibv_cq->read_vendor_err()`; making
  * a best-effort attempt to promote the error code to a proprietary EFA
  * provider error code.
  *
@@ -384,8 +547,8 @@ static void efa_rdm_cq_handle_recv_completion(struct efa_ibv_cq *ibv_cq, struct
  * RDMA Core error codes (#EFA_IO_COMP_STATUSES) for the sake of more accurate
  * error reporting
  */
-static int efa_rdm_cq_get_prov_errno(struct ibv_cq_ex *ibv_cq_ex, struct efa_rdm_peer *peer) {
-	uint32_t vendor_err = ibv_wc_read_vendor_err(ibv_cq_ex);
+static int efa_rdm_cq_get_prov_errno(struct efa_ibv_cq *ibv_cq, struct efa_rdm_peer *peer) {
+	uint32_t vendor_err = efa_ibv_cq_wc_read_vendor_err(ibv_cq);
 
 	if (OFI_UNLIKELY(!peer)) {
 		return vendor_err;
@@ -410,104 +573,158 @@ static int efa_rdm_cq_match_ep(struct dlist_entry *item, const void *ep)
 	return (container_of(item, struct efa_rdm_ep, entry) == ep) ;
 }
 
-/**
- * @brief poll rdma-core cq and process the cq entry
- *
- * @param[in]	ep_poll	the RDM endpoint that polls ibv cq. Note this polling endpoint can be different
- * from the endpoint that the completed packet entry was posted from (pkt_entry->ep).
- * @param[in]	cqe_to_process	Max number of cq entry to poll and process. A negative number means to poll until cq empty
- */
-void efa_rdm_cq_poll_ibv_cq(ssize_t cqe_to_process, struct efa_ibv_cq *ibv_cq)
+static inline struct efa_rdm_ep *efa_rdm_cq_get_rdm_ep(struct efa_ibv_cq *cq, struct efa_domain *efa_domain)
 {
-	bool should_end_poll = false;
-	/* Initialize an empty ibv_poll_cq_attr struct for ibv_start_poll.
-	 * EFA expects .comp_mask = 0, or otherwise returns EINVAL.
-	 */
-	struct ibv_poll_cq_attr poll_cq_attr = {.comp_mask = 0};
-	struct efa_rdm_pke *pkt_entry;
-	ssize_t err;
-	int opcode;
-	size_t i = 0;
+	struct efa_base_ep *base_ep = efa_domain->qp_table[efa_ibv_cq_wc_read_qp_num(cq) & efa_domain->qp_table_sz_m1]->base_ep;
+	return container_of(base_ep, struct efa_rdm_ep, base_ep);
+}
+
+/**
+ * @brief Process work completions for a closing endpoint
+ *
+ * This is a lighter-weight counterpart to #efa_rdm_cq_process_wc(); avoiding
+ * unnecessary overhead for processing completions for an endpoint that's
+ * closing anyway by simply releasing the packet entries. Exceptions include
+ * RECEIPT and EOR packets when a completion fails due to RNR and resource
+ * management is enabled by the user (FI_RM_ENABLED). In this case, packets are
+ * queued to be reposted.
+ *
+ * @param[in]	cq	IBV CQ
+ * @param[in]	ep	EFA RDM endpoint (to be closed)
+ * @return	Status code for the work completion
+ */
+static inline
+enum ibv_wc_status efa_rdm_cq_process_wc_closing_ep(struct efa_ibv_cq *cq, struct efa_rdm_ep *ep)
+{
+	uint64_t wr_id = cq->ibv_cq_ex->wr_id;
+	enum ibv_wc_status status = cq->ibv_cq_ex->status;
+	enum ibv_wc_opcode opcode = efa_ibv_cq_wc_read_opcode(cq);
+	struct efa_rdm_pke *pkt_entry = (struct efa_rdm_pke *) wr_id;
 	int prov_errno;
-	struct efa_rdm_ep *ep = NULL;
-	struct fi_cq_err_entry err_entry;
-	struct efa_cq *efa_cq;
-	struct efa_domain *efa_domain;
-	struct efa_qp *qp;
-	struct efa_rdm_peer *peer = NULL;
-	struct dlist_entry rx_progressed_ep_list, *tmp;
 
-	efa_cq = container_of(ibv_cq, struct efa_cq, ibv_cq);
-	efa_domain = container_of(efa_cq->util_cq.domain, struct efa_domain, util_domain);
-	dlist_init(&rx_progressed_ep_list);
-
-	/* Call ibv_start_poll only once */
-	err = ibv_start_poll(ibv_cq->ibv_cq_ex, &poll_cq_attr);
-	should_end_poll = !err;
-
-	while (!err) {
-		pkt_entry = (void *)(uintptr_t)ibv_cq->ibv_cq_ex->wr_id;
-		qp = efa_domain->qp_table[ibv_wc_read_qp_num(ibv_cq->ibv_cq_ex) & efa_domain->qp_table_sz_m1];
-		ep = container_of(qp->base_ep, struct efa_rdm_ep, base_ep);
 #if HAVE_LTTNG
-		efa_rdm_tracepoint(poll_cq, (size_t) ibv_cq->ibv_cq_ex->wr_id);
-		if (pkt_entry && pkt_entry->ope)
-			efa_rdm_tracepoint(poll_cq_ope, pkt_entry->ope->msg_id,
-					   (size_t) pkt_entry->ope->cq_entry.op_context,
-					   pkt_entry->ope->total_len, pkt_entry->ope->cq_entry.tag,
-					   pkt_entry->ope->addr);
+	efa_rdm_tracepoint(poll_cq, (size_t) wr_id);
+	if (pkt_entry && pkt_entry->ope)
+		efa_rdm_tracepoint(poll_cq_ope, pkt_entry->ope->msg_id,
+				   (size_t) pkt_entry->ope->cq_entry.op_context,
+				   pkt_entry->ope->total_len, pkt_entry->ope->cq_entry.tag,
+				   pkt_entry->ope->peer ? pkt_entry->ope->peer->conn->fi_addr : FI_ADDR_NOTAVAIL);
 #endif
-		opcode = ibv_wc_read_opcode(ibv_cq->ibv_cq_ex);
-		if (ibv_cq->ibv_cq_ex->status) {
-			if (pkt_entry)
-				peer = efa_rdm_ep_get_peer(ep, pkt_entry->addr);
-			prov_errno = efa_rdm_cq_get_prov_errno(ibv_cq->ibv_cq_ex, peer);
-			switch (opcode) {
-			case IBV_WC_SEND: /* fall through */
-			case IBV_WC_RDMA_WRITE: /* fall through */
-			case IBV_WC_RDMA_READ:
-				efa_rdm_pke_handle_tx_error(pkt_entry, prov_errno, peer);
-				break;
-			case IBV_WC_RECV: /* fall through */
-			case IBV_WC_RECV_RDMA_WITH_IMM:
-				if (efa_cq_wc_is_unsolicited(ibv_cq->ibv_cq_ex)) {
-					EFA_WARN(FI_LOG_CQ, "Receive error %s (%d) for unsolicited write recv",
-						efa_strerror(prov_errno), prov_errno);
-					efa_base_ep_write_eq_error(&ep->base_ep, to_fi_errno(prov_errno), prov_errno);
+
+	if (!efa_cq_wc_is_unsolicited(cq)) {
+		if (OFI_UNLIKELY(status != IBV_WC_SUCCESS)) {
+			prov_errno = efa_rdm_cq_get_prov_errno(cq, pkt_entry->peer);
+			if (prov_errno == EFA_IO_COMP_STATUS_REMOTE_ERROR_RNR &&
+				ep->handle_resource_management == FI_RM_ENABLED) {
+				switch(efa_rdm_pkt_type_of(pkt_entry)) {
+				case EFA_RDM_RECEIPT_PKT:
+				case EFA_RDM_EOR_PKT:
+					efa_rdm_ep_record_tx_op_completed(ep, pkt_entry);
+					efa_rdm_ep_queue_rnr_pkt(ep, pkt_entry);
+					return status;
+				default:
 					break;
 				}
-				efa_rdm_pke_handle_rx_error(pkt_entry, prov_errno);
-				break;
-			default:
-				EFA_WARN(FI_LOG_EP_CTRL, "Unhandled op code %d\n", opcode);
-				assert(0 && "Unhandled op code");
 			}
-			break;
 		}
+		switch (opcode) {
+		case IBV_WC_SEND: /* fall through */
+		case IBV_WC_RDMA_WRITE: /* fall through */
+		case IBV_WC_RDMA_READ:
+			efa_rdm_ep_record_tx_op_completed(ep, pkt_entry);
+			efa_rdm_pke_release_tx(pkt_entry);
+			break;
+		case IBV_WC_RECV: /* fall through */
+		case IBV_WC_RECV_RDMA_WITH_IMM:
+			if (pkt_entry->alloc_type == EFA_RDM_PKE_FROM_USER_RX_POOL) {
+				assert(ep->user_rx_pkts_posted > 0);
+				ep->user_rx_pkts_posted--;
+			} else {
+				assert(ep->efa_rx_pkts_posted > 0);
+				ep->efa_rx_pkts_posted--;
+			}
+			efa_rdm_pke_release_rx(pkt_entry);
+			break;
+		default:
+			EFA_WARN(FI_LOG_EP_CTRL, "Unhandled opcode: %d\n", opcode);
+			assert(0 && "Unhandled opcode");
+		}
+	}
+	return status;
+}
+
+/**
+ * @brief Process work completions
+ *
+ * @param[in]	cq	IBV CQ
+ * @param[in]	ep	EFA RDM endpoint
+ * @return	Status code for the work completion
+ */
+static inline
+enum ibv_wc_status efa_rdm_cq_process_wc(struct efa_ibv_cq *cq, struct efa_rdm_ep *ep)
+{
+	uint64_t wr_id = cq->ibv_cq_ex->wr_id;
+	enum ibv_wc_status status = cq->ibv_cq_ex->status;
+	enum ibv_wc_opcode opcode = efa_ibv_cq_wc_read_opcode(cq);
+	struct efa_rdm_pke *pkt_entry = (struct efa_rdm_pke *) wr_id;
+	int prov_errno;
+
+#if HAVE_LTTNG
+	efa_rdm_tracepoint(poll_cq, (size_t) wr_id);
+	if (pkt_entry && pkt_entry->ope)
+		efa_rdm_tracepoint(poll_cq_ope, pkt_entry->ope->msg_id,
+				   (size_t) pkt_entry->ope->cq_entry.op_context,
+				   pkt_entry->ope->total_len, pkt_entry->ope->cq_entry.tag,
+				   pkt_entry->ope->peer ? pkt_entry->ope->peer->conn->fi_addr : FI_ADDR_NOTAVAIL);
+#endif
+
+	if (OFI_UNLIKELY(status != IBV_WC_SUCCESS)) {
+		prov_errno = efa_rdm_cq_get_prov_errno(cq, pkt_entry ? pkt_entry->peer : NULL);
+		switch (opcode) {
+		case IBV_WC_SEND: /* fall through */
+		case IBV_WC_RDMA_WRITE: /* fall through */
+		case IBV_WC_RDMA_READ:
+			assert(pkt_entry);
+			efa_rdm_pke_handle_tx_error(pkt_entry, prov_errno);
+			break;
+		case IBV_WC_RECV: /* fall through */
+		case IBV_WC_RECV_RDMA_WITH_IMM:
+			if (efa_cq_wc_is_unsolicited(cq)) {
+				EFA_WARN(FI_LOG_CQ, "Receive error %s (%d) for unsolicited write recv",
+					efa_strerror(prov_errno), prov_errno);
+				efa_base_ep_write_eq_error(&ep->base_ep, to_fi_errno(prov_errno), prov_errno);
+				break;
+			}
+			assert(pkt_entry);
+			efa_rdm_pke_handle_rx_error(pkt_entry, prov_errno);
+			break;
+		default:
+			EFA_WARN(FI_LOG_EP_CTRL, "Unhandled opcode: %d\n", opcode);
+			assert(0 && "Unhandled opcode");
+		}
+	} else {
 		switch (opcode) {
 		case IBV_WC_SEND:
 #if ENABLE_DEBUG
 			ep->send_comps++;
 #endif
-			peer = efa_rdm_ep_get_peer(ep, pkt_entry->addr);
-			efa_rdm_pke_handle_send_completion(pkt_entry, peer);
+			efa_rdm_pke_handle_send_completion(pkt_entry);
 			break;
 		case IBV_WC_RECV:
 			/* efa_rdm_cq_handle_recv_completion does additional work to determine the source
 			 * address and the peer struct. So do not try to identify the peer here. */
-			efa_rdm_cq_handle_recv_completion(ibv_cq, pkt_entry, ep);
+			efa_rdm_cq_handle_recv_completion(cq, pkt_entry, ep);
 #if ENABLE_DEBUG
 			ep->recv_comps++;
 #endif
 			break;
 		case IBV_WC_RDMA_READ:
 		case IBV_WC_RDMA_WRITE:
-			peer = efa_rdm_ep_get_peer(ep, pkt_entry->addr);
-			efa_rdm_pke_handle_rma_completion(pkt_entry, peer);
+			efa_rdm_pke_handle_rma_completion(pkt_entry);
 			break;
 		case IBV_WC_RECV_RDMA_WITH_IMM:
 			efa_rdm_cq_proc_ibv_recv_rdma_with_imm_completion(
-				ibv_cq->ibv_cq_ex,
+				cq,
 				FI_REMOTE_CQ_DATA | FI_RMA | FI_REMOTE_WRITE,
 				ep, pkt_entry);
 			break;
@@ -516,37 +733,83 @@ void efa_rdm_cq_poll_ibv_cq(ssize_t cqe_to_process, struct efa_ibv_cq *ibv_cq)
 				"Unhandled cq type\n");
 			assert(0 && "Unhandled cq type");
 		}
+	}
+	return status;
+}
 
+void efa_rdm_cq_poll_ibv_cq_closing_ep(struct efa_ibv_cq *ibv_cq, struct efa_rdm_ep *closing_ep)
+{
+
+	struct efa_rdm_ep *ep = NULL;
+	struct efa_cq *efa_cq = container_of(ibv_cq, struct efa_cq, ibv_cq);
+	struct efa_domain *efa_domain = container_of(efa_cq->util_cq.domain, struct efa_domain, util_domain);
+	struct dlist_entry rx_progressed_ep_list, *tmp;
+
+	dlist_init(&rx_progressed_ep_list);
+
+	efa_cq_start_poll(ibv_cq);
+	while (efa_cq_wc_available(ibv_cq)) {
+		ep = efa_rdm_cq_get_rdm_ep(ibv_cq, efa_domain);
+		if (ep == closing_ep) {
+			if (OFI_UNLIKELY(efa_rdm_cq_process_wc_closing_ep(ibv_cq, ep) != IBV_WC_SUCCESS))
+				break;
+		} else {
+			if (OFI_UNLIKELY(efa_rdm_cq_process_wc(ibv_cq, ep) != IBV_WC_SUCCESS))
+				break;
+			if (ep->efa_rx_pkts_to_post > 0 && !dlist_find_first_match(&rx_progressed_ep_list, &efa_rdm_cq_match_ep, ep))
+				dlist_insert_tail(&ep->entry, &rx_progressed_ep_list);
+		}
+		efa_cq_next_poll(ibv_cq);
+	}
+	efa_cq_end_poll(ibv_cq);
+	dlist_foreach_container_safe(
+		&rx_progressed_ep_list, struct efa_rdm_ep, ep, entry, tmp) {
+		efa_rdm_ep_post_internal_rx_pkts(ep);
+		dlist_remove(&ep->entry);
+	}
+	assert(dlist_empty(&rx_progressed_ep_list));
+}
+
+/**
+ * @brief poll rdma-core cq and process the cq entry
+ *
+ * @param[in]	ep_poll	the RDM endpoint that polls ibv cq. Note this polling endpoint can be different
+ * from the endpoint that the completed packet entry was posted from (pkt_entry->ep).
+ * @param[in]	cqe_to_process	Max number of cq entry to poll and process. A negative number means to poll until cq empty
+ */
+int efa_rdm_cq_poll_ibv_cq(ssize_t cqe_to_process, struct efa_ibv_cq *ibv_cq)
+{
+	int err;
+	size_t i = 0;
+	struct efa_rdm_ep *ep = NULL;
+	struct efa_cq *efa_cq = container_of(ibv_cq, struct efa_cq, ibv_cq);
+	struct efa_domain *efa_domain = container_of(efa_cq->util_cq.domain, struct efa_domain, util_domain);
+
+	struct dlist_entry rx_progressed_ep_list, *tmp;
+
+	dlist_init(&rx_progressed_ep_list);
+
+	/* Call ibv_start_poll only once */
+	efa_cq_start_poll(ibv_cq);
+
+	while (efa_cq_wc_available(ibv_cq)) {
+		ep = efa_rdm_cq_get_rdm_ep(ibv_cq, efa_domain);
+		if (OFI_UNLIKELY(efa_rdm_cq_process_wc(ibv_cq, ep) != IBV_WC_SUCCESS))
+			break;
 		if (ep->efa_rx_pkts_to_post > 0 && !dlist_find_first_match(&rx_progressed_ep_list, &efa_rdm_cq_match_ep, ep))
 			dlist_insert_tail(&ep->entry, &rx_progressed_ep_list);
-		i++;
-		if (i == cqe_to_process) {
+		if (++i >= cqe_to_process)
 			break;
-		}
 
 		/*
 		 * ibv_next_poll MUST be call after the current WC is fully processed,
 		 * which prevents later calls on ibv_cq_ex from reading the wrong WC.
 		 */
-		err = ibv_next_poll(ibv_cq->ibv_cq_ex);
+		efa_cq_next_poll(ibv_cq);
 	}
 
-	if (err && err != ENOENT) {
-		err = err > 0 ? err : -err;
-		prov_errno = ibv_wc_read_vendor_err(ibv_cq->ibv_cq_ex);
-		EFA_WARN(FI_LOG_CQ, "Unexpected error when polling ibv cq, err: %s (%zd) prov_errno: %s (%d)\n", fi_strerror(err), err, efa_strerror(prov_errno), prov_errno);
-		efa_show_help(prov_errno);
-		err_entry = (struct fi_cq_err_entry) {
-			.err = err,
-			.prov_errno = prov_errno,
-			.op_context = NULL
-		};
-		ofi_cq_write_error(&efa_cq->util_cq, &err_entry);
-	}
-
-	if (should_end_poll)
-		ibv_end_poll(ibv_cq->ibv_cq_ex);
-
+	err = ibv_cq->poll_err;
+	efa_cq_end_poll(ibv_cq);
 	dlist_foreach_container_safe(
 		&rx_progressed_ep_list, struct efa_rdm_ep, ep, entry, tmp) {
 		efa_rdm_ep_post_internal_rx_pkts(ep);
@@ -554,6 +817,7 @@ void efa_rdm_cq_poll_ibv_cq(ssize_t cqe_to_process, struct efa_ibv_cq *ibv_cq)
 	}
 	assert(dlist_empty(&rx_progressed_ep_list));
 
+	return err;
 }
 
 static ssize_t efa_rdm_cq_readfrom(struct fid_cq *cq_fid, void *buf, size_t count, fi_addr_t *src_addr)
@@ -633,7 +897,7 @@ static void efa_rdm_cq_progress(struct util_cq *cq)
 
 	dlist_foreach(&efa_rdm_cq->ibv_cq_poll_list, item) {
 		poll_list_entry = container_of(item, struct efa_ibv_cq_poll_list_entry, entry);
-		efa_rdm_cq_poll_ibv_cq(efa_env.efa_cq_read_size, poll_list_entry->cq);
+		(void) efa_rdm_cq_poll_ibv_cq(efa_env.efa_cq_read_size, poll_list_entry->cq);
 	}
 	efa_domain_progress_rdm_peers_and_queues(efa_domain);
 	ofi_genlock_unlock(&cq->ep_list_lock);
@@ -682,14 +946,16 @@ int efa_rdm_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 	if (ret)
 		goto free;
 
-	ret = efa_cq_ibv_cq_ex_open(
-		attr, efa_domain->device->ibv_ctx, &cq->efa_cq.ibv_cq.ibv_cq_ex,
-		&cq->efa_cq.ibv_cq.ibv_cq_ex_type, &efa_cq_init_attr);
+	ret = efa_cq_open_ibv_cq(
+		attr, efa_domain->device->ibv_ctx, &cq->efa_cq.ibv_cq,
+		&efa_cq_init_attr);
 	if (ret) {
 		EFA_WARN(FI_LOG_CQ, "Unable to create extended CQ: %s\n", fi_strerror(ret));
 		goto close_util_cq;
 	}
 
+	cq->efa_cq.poll_ibv_cq = efa_rdm_cq_poll_ibv_cq;
+
 	*cq_fid = &cq->efa_cq.util_cq.cq_fid;
 	(*cq_fid)->fid.ops = &efa_rdm_cq_fi_ops;
 	(*cq_fid)->ops = &efa_rdm_cq_ops;
