@@ -12,6 +12,7 @@
 #include "efa_rdm_rxe_map.h"
 #include "efa_rdm_pkt_type.h"
 #include "efa_rdm_pke_req.h"
+#include "efa_rdm_pke_utils.h"
 #include "efa_cntr.h"
 
 
@@ -237,12 +238,28 @@ int efa_rdm_ep_create_buffer_pools(struct efa_rdm_ep *ep)
 				 sizeof(struct efa_rdm_ep_peer_map_entry),
 				 EFA_RDM_BUFPOOL_ALIGNMENT, 0, /* no limit to max_cnt */
 				 EFA_RDM_EP_MIN_PEER_POOL_SIZE,
-				 /* Don't track usage, because endpoint can be closed without removing entries from AV */
-				 OFI_BUFPOOL_NO_TRACK);
+				 0);
+	if (ret)
+		goto err_free;
+
+	ret = ofi_bufpool_grow(ep->peer_map_entry_pool);
+	if (ret)
+		goto err_free;
+
+	ret = ofi_bufpool_create(&ep->peer_robuf_pool,
+				(sizeof(struct efa_rdm_pke*) * (roundup_power_of_two(efa_env.recvwin_size)) +
+				sizeof(struct recvwin_cirq)),
+				EFA_RDM_BUFPOOL_ALIGNMENT, 0, /* no limit to max_cnt */
+				EFA_RDM_EP_MIN_PEER_REORDER_BUFFER_POOL_SIZE,
+				0);
+
+	if (ret)
+		goto err_free;
+
+	ret = ofi_bufpool_grow(ep->peer_robuf_pool);
 	if (ret)
 		goto err_free;
 
-	efa_rdm_rxe_map_construct(&ep->rxe_map);
 	return 0;
 
 err_free:
@@ -279,6 +296,9 @@ err_free:
 	if (ep->peer_map_entry_pool)
 		ofi_bufpool_destroy(ep->peer_map_entry_pool);
 
+	if (ep->peer_robuf_pool)
+		ofi_bufpool_destroy(ep->peer_robuf_pool);
+
 	return ret;
 }
 
@@ -479,10 +499,16 @@ int efa_rdm_ep_open(struct fid_domain *domain, struct fi_info *info,
 		goto err_free_ep;
 
 	if (efa_domain->shm_domain) {
-		ret = fi_endpoint(efa_domain->shm_domain, efa_domain->shm_info,
-				  &efa_rdm_ep->shm_ep, efa_rdm_ep);
-		if (ret)
-			goto err_destroy_base_ep;
+		efa_rdm_ep->shm_info = NULL;
+		efa_shm_info_create(info, &efa_rdm_ep->shm_info);
+		if (efa_rdm_ep->shm_info) {
+			ret = fi_endpoint(efa_domain->shm_domain, efa_rdm_ep->shm_info,
+					  &efa_rdm_ep->shm_ep, efa_rdm_ep);
+			if (ret)
+				goto err_destroy_base_ep;
+		} else {
+			efa_rdm_ep->shm_ep = NULL;
+		}
 	} else {
 		efa_rdm_ep->shm_ep = NULL;
 	}
@@ -547,11 +573,6 @@ int efa_rdm_ep_open(struct fid_domain *domain, struct fi_info *info,
 	efa_rdm_ep->efa_rx_pkts_held = 0;
 	efa_rdm_ep->efa_outstanding_tx_ops = 0;
 
-	if (ret) {
-		EFA_WARN(FI_LOG_CQ, "Unable to create extended CQ: %s\n", strerror(errno));
-		goto err_close_shm_ep;
-	}
-
 	ret = efa_rdm_ep_create_buffer_pools(efa_rdm_ep);
 	if (ret)
 		goto err_close_shm_ep;
@@ -711,7 +732,6 @@ static void efa_rdm_ep_destroy_buffer_pools(struct efa_rdm_ep *efa_rdm_ep)
 	struct efa_rdm_ope *rxe;
 	struct efa_rdm_ope *txe;
 
-	ofi_genlock_lock(&efa_rdm_ep_domain(efa_rdm_ep)->srx_lock);
 #if ENABLE_DEBUG
 	struct efa_rdm_pke *pkt_entry;
 
@@ -754,7 +774,24 @@ static void efa_rdm_ep_destroy_buffer_pools(struct efa_rdm_ep *efa_rdm_ep)
 		efa_rdm_txe_release(txe);
 	}
 
-	ofi_genlock_unlock(&efa_rdm_ep_domain(efa_rdm_ep)->srx_lock);
+	/* Clean up any remaining peers in the hashmap before destroying buffer pools */
+	if (efa_rdm_ep->fi_addr_to_peer_map) {
+		struct efa_rdm_ep_peer_map_entry *map_entry, *map_tmp;
+		HASH_ITER(hndl, efa_rdm_ep->fi_addr_to_peer_map, map_entry, map_tmp) {
+			efa_rdm_peer_destruct(&map_entry->peer, efa_rdm_ep);
+			HASH_DELETE(hndl, efa_rdm_ep->fi_addr_to_peer_map, map_entry);
+			ofi_buf_free(map_entry);
+		}
+	}
+
+	if (efa_rdm_ep->fi_addr_to_peer_map_implicit) {
+		struct efa_rdm_ep_peer_map_entry *map_entry, *map_tmp;
+		HASH_ITER(hndl, efa_rdm_ep->fi_addr_to_peer_map_implicit, map_entry, map_tmp) {
+			efa_rdm_peer_destruct(&map_entry->peer, efa_rdm_ep);
+			HASH_DELETE(hndl, efa_rdm_ep->fi_addr_to_peer_map_implicit, map_entry);
+			ofi_buf_free(map_entry);
+		}
+	}
 
 	if (efa_rdm_ep->ope_pool)
 		ofi_bufpool_destroy(efa_rdm_ep->ope_pool);
@@ -794,6 +831,9 @@ static void efa_rdm_ep_destroy_buffer_pools(struct efa_rdm_ep *efa_rdm_ep)
 
 	if (efa_rdm_ep->peer_map_entry_pool)
 		ofi_bufpool_destroy(efa_rdm_ep->peer_map_entry_pool);
+
+	if (efa_rdm_ep->peer_robuf_pool)
+		ofi_bufpool_destroy(efa_rdm_ep->peer_robuf_pool);
 }
 
 /*
@@ -805,7 +845,6 @@ static void efa_rdm_ep_destroy_buffer_pools(struct efa_rdm_ep *efa_rdm_ep)
  * @param[in]  efa_rdm_ep      endpoint
  * @return     a boolean
  */
-static
 bool efa_rdm_ep_has_unfinished_send(struct efa_rdm_ep *efa_rdm_ep)
 {
 	struct dlist_entry *entry, *tmp;
@@ -827,6 +866,48 @@ bool efa_rdm_ep_has_unfinished_send(struct efa_rdm_ep *efa_rdm_ep)
     return false;
 }
 
+static inline void progress_queues_closing_ep(struct efa_rdm_ep *ep)
+{
+	struct efa_rdm_peer *peer;
+	struct dlist_entry *tmp;
+	struct efa_rdm_ope *ope;
+	struct efa_domain *domain = efa_rdm_ep_domain(ep);
+
+	assert(domain->info->ep_attr->type == FI_EP_RDM);
+
+	/* Update timers for peers that are in backoff list*/
+	dlist_foreach_container_safe(&domain->peer_backoff_list,
+			struct efa_rdm_peer, peer, rnr_backoff_entry, tmp) {
+		if (ofi_gettime_us() >= peer->rnr_backoff_begin_ts +
+					peer->rnr_backoff_wait_time) {
+			peer->flags &= ~EFA_RDM_PEER_IN_BACKOFF;
+			dlist_remove(&peer->rnr_backoff_entry);
+		}
+	}
+
+	dlist_foreach_container_safe(&domain->ope_queued_list,
+			struct efa_rdm_ope, ope, queued_entry, tmp) {
+		if (ope->ep == ep) {
+			switch (efa_rdm_pke_get_ctrl_pkt_type_from_queued_ope(ope)) {
+			case EFA_RDM_RECEIPT_PKT:
+			case EFA_RDM_EOR_PKT:
+				if (efa_rdm_ope_process_queued_ope(ope, EFA_RDM_OPE_QUEUED_RNR))
+					continue;
+				if (efa_rdm_ope_process_queued_ope(ope, EFA_RDM_OPE_QUEUED_CTRL))
+					continue;
+				break;
+			default:
+				/* Release all other queued OPEs */
+				if (ope->type == EFA_RDM_TXE)
+					efa_rdm_txe_release(ope);
+				else
+					efa_rdm_rxe_release(ope);
+				break;
+			}
+		}
+	}
+}
+
 /*
  * @brief wait for send to finish
  *
@@ -849,10 +930,10 @@ void efa_rdm_ep_wait_send(struct efa_rdm_ep *efa_rdm_ep)
 	while (efa_rdm_ep_has_unfinished_send(efa_rdm_ep)) {
 		/* poll cq until empty */
 		if (tx_cq)
-			efa_rdm_cq_poll_ibv_cq(-1, &tx_cq->ibv_cq);
+			efa_rdm_cq_poll_ibv_cq_closing_ep(&tx_cq->ibv_cq, efa_rdm_ep);
 		if (rx_cq)
-			efa_rdm_cq_poll_ibv_cq(-1, &rx_cq->ibv_cq);
-		efa_domain_progress_rdm_peers_and_queues(efa_rdm_ep_domain(efa_rdm_ep));
+			efa_rdm_cq_poll_ibv_cq_closing_ep(&rx_cq->ibv_cq, efa_rdm_ep);
+		progress_queues_closing_ep(efa_rdm_ep);
 	}
 
 	ofi_genlock_unlock(&efa_rdm_ep_domain(efa_rdm_ep)->srx_lock);
@@ -916,6 +997,11 @@ static int efa_rdm_ep_close_shm_ep_resources(struct efa_rdm_ep *efa_rdm_ep)
 		efa_rdm_ep->shm_ep = NULL;
 	}
 
+	if (efa_rdm_ep->shm_info) {
+		fi_freeinfo(efa_rdm_ep->shm_info);
+		efa_rdm_ep->shm_info = NULL;
+	}
+
 	return retv;
 }
 
@@ -930,7 +1016,6 @@ static int efa_rdm_ep_close(struct fid *fid)
 	struct efa_domain *domain;
 	struct dlist_entry *entry, *tmp;
 	struct efa_rdm_ope *rxe;
-	struct efa_rdm_peer *peer;
 
 	efa_rdm_ep = container_of(fid, struct efa_rdm_ep, base_ep.util_ep.ep_fid.fid);
 	domain = efa_rdm_ep_domain(efa_rdm_ep);
@@ -964,22 +1049,12 @@ static int efa_rdm_ep_close(struct fid *fid)
 		efa_rdm_ep->peer_srx_ep = NULL;
 	}
 
-	/* Remove all peers associated with this endpoint in domain level peer lists */
+	/**
+	 * The QP destroy and op entries clean up must be in the same lock,
+	 * otherwise there can be race condition that efa_domain_progress_rdm_peers_and_queues
+	 * (part of fi_cq_read) can access entries that are from a closed QP.
+	 */
 	ofi_genlock_lock(&domain->srx_lock);
-	dlist_foreach_container_safe(&domain->peer_backoff_list, struct efa_rdm_peer,
-				     peer, rnr_backoff_entry, tmp) {
-		if (peer->ep == efa_rdm_ep) {
-			dlist_remove(&peer->rnr_backoff_entry);
-		}
-	}
-
-	dlist_foreach_container_safe(&domain->handshake_queued_peer_list, struct efa_rdm_peer,
-				     peer, handshake_queued_entry, tmp) {
-		if (peer->ep == efa_rdm_ep) {
-			dlist_remove(&peer->handshake_queued_entry);
-		}
-	}
-	ofi_genlock_unlock(&domain->srx_lock);
 
 	/* We need to free the util_ep first to avoid race conditions
 	 * with other threads progressing the cq. */
@@ -1004,6 +1079,8 @@ static int efa_rdm_ep_close(struct fid *fid)
 	if (efa_rdm_ep->pke_vec)
 		free(efa_rdm_ep->pke_vec);
 
+	ofi_genlock_unlock(&domain->srx_lock);
+
 	free(efa_rdm_ep);
 	return retv;
 }
@@ -1032,7 +1109,7 @@ void efa_rdm_ep_set_extra_info(struct efa_rdm_ep *ep)
 
 	ep->extra_info[0] |= EFA_RDM_EXTRA_FEATURE_DELIVERY_COMPLETE;
 
-	if (efa_use_unsolicited_write_recv())
+	if (ep->base_ep.qp->unsolicited_write_recv_enabled)
 		ep->extra_info[0] |= EFA_RDM_EXTRA_FEATURE_UNSOLICITED_WRITE_RECV;
 
 	if (ep->use_zcpy_rx) {
@@ -1108,11 +1185,6 @@ static void efa_rdm_ep_close_shm_resources(struct efa_rdm_ep *efa_rdm_ep)
 			EFA_WARN(FI_LOG_EP_CTRL, "Unable to close shm fabric\n");
 		efa_domain->fabric->shm_fabric = NULL;
 	}
-
-	if (efa_domain->shm_info) {
-		fi_freeinfo(efa_domain->shm_info);
-		efa_domain->shm_info = NULL;
-	}
 }
 
 /**
