@@ -26,7 +26,6 @@ void efa_rdm_txe_construct(struct efa_rdm_ope *txe,
 	txe->op = op;
 	txe->tx_id = ofi_buf_index(txe);
 	txe->state = EFA_RDM_TXE_REQ;
-	txe->addr = msg->addr;
 	txe->peer = peer;
 	/* peer would be NULL for local read operation */
 	if (txe->peer) {
@@ -148,14 +147,15 @@ void efa_rdm_txe_release(struct efa_rdm_ope *txe)
 		efa_rdm_pke_release_tx(pkt_entry);
 	}
 
-	if (txe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS)
+	if (txe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS) {
 		dlist_remove(&txe->queued_entry);
+		txe->internal_flags &= ~EFA_RDM_OPE_QUEUED_FLAGS;
+	}
 
 #ifdef ENABLE_EFA_POISONING
 	efa_rdm_poison_mem_region(txe,
 			      sizeof(struct efa_rdm_ope));
 #endif
-	txe->state = EFA_RDM_OPE_FREE;
 	ofi_buf_free(txe);
 }
 
@@ -184,7 +184,7 @@ void efa_rdm_rxe_release_internal(struct efa_rdm_ope *rxe)
 		dlist_remove(&rxe->entry);
 
 	if (rxe->rxe_map)
-		efa_rdm_rxe_map_remove(rxe->rxe_map, rxe->msg_id, rxe->addr, rxe);
+		efa_rdm_rxe_map_remove(rxe->rxe_map, rxe->msg_id, rxe);
 
 	for (i = 0; i < rxe->iov_count; i++) {
 		if (rxe->mr[i]) {
@@ -203,14 +203,15 @@ void efa_rdm_rxe_release_internal(struct efa_rdm_ope *rxe)
 				     pkt_entry, entry, tmp)
 		efa_rdm_pke_release_tx(pkt_entry);
 
-	if (rxe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS)
+	if (rxe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS) {
 		dlist_remove(&rxe->queued_entry);
+		rxe->internal_flags &= ~EFA_RDM_OPE_QUEUED_FLAGS;
+	}
 
 #ifdef ENABLE_EFA_POISONING
 	efa_rdm_poison_mem_region(rxe,
 			      sizeof(struct efa_rdm_ope));
 #endif
-	rxe->state = EFA_RDM_OPE_FREE;
 	ofi_buf_free(rxe);
 }
 
@@ -600,24 +601,34 @@ void efa_rdm_rxe_handle_error(struct efa_rdm_ope *rxe, int err, int prov_errno)
 	case EFA_RDM_RXE_UNEXP:
 	case EFA_RDM_RXE_MATCHED:
 		break;
+	case EFA_RDM_OPE_SEND:
+		dlist_remove(&rxe->entry);
+		break;
 	case EFA_RDM_RXE_RECV:
 #if ENABLE_DEBUG
 		dlist_remove(&rxe->pending_recv_entry);
 #endif
 		break;
+	case EFA_RDM_OPE_ERR:
+		/* Already progressed, no-op */
+		return;
 	default:
 		EFA_WARN(FI_LOG_CQ, "rxe unknown state %d\n",
 			rxe->state);
 		assert(0 && "rxe unknown state");
 	}
 
+	rxe->state = EFA_RDM_OPE_ERR;
+
 	dlist_foreach_container_safe(&rxe->queued_pkts,
 				     struct efa_rdm_pke,
 				     pkt_entry, entry, tmp)
 		efa_rdm_pke_release_tx(pkt_entry);
 
-	if (rxe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS)
+	if (rxe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS) {
 		dlist_remove(&rxe->queued_entry);
+		rxe->internal_flags &= ~EFA_RDM_OPE_QUEUED_FLAGS;
+	}
 
 	if (rxe->unexp_pkt) {
 		efa_rdm_pke_release_rx_list(rxe->unexp_pkt);
@@ -630,7 +641,7 @@ void efa_rdm_rxe_handle_error(struct efa_rdm_ope *rxe, int err, int prov_errno)
 	err_entry.buf = rxe->cq_entry.buf;
 	err_entry.data = rxe->cq_entry.data;
 	err_entry.tag = rxe->cq_entry.tag;
-	if (OFI_UNLIKELY(efa_rdm_write_error_msg(ep, rxe->addr, prov_errno,
+	if (OFI_UNLIKELY(efa_rdm_write_error_msg(ep, rxe->peer, prov_errno,
 	                                         err_msg, &err_entry.err_data_size))) {
 		err_entry.err_data_size = 0;
 	} else {
@@ -713,14 +724,21 @@ void efa_rdm_txe_handle_error(struct efa_rdm_ope *txe, int err, int prov_errno)
 	case EFA_RDM_OPE_SEND:
 		dlist_remove(&txe->entry);
 		break;
+	case EFA_RDM_OPE_ERR:
+		/* Already progressed, no-op */
+		return;
 	default:
 		EFA_WARN(FI_LOG_CQ, "txe unknown state %d\n",
 			txe->state);
 		assert(0 && "txe unknown state");
 	}
 
-	if (txe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS)
+	txe->state = EFA_RDM_OPE_ERR;
+
+	if (txe->internal_flags & EFA_RDM_OPE_QUEUED_FLAGS) {
 		dlist_remove(&txe->queued_entry);
+		txe->internal_flags &= ~EFA_RDM_OPE_QUEUED_FLAGS;
+	}
 
 	dlist_foreach_container_safe(&txe->queued_pkts,
 				     struct efa_rdm_pke,
@@ -732,7 +750,7 @@ void efa_rdm_txe_handle_error(struct efa_rdm_ope *txe, int err, int prov_errno)
 	err_entry.buf = txe->cq_entry.buf;
 	err_entry.data = txe->cq_entry.data;
 	err_entry.tag = txe->cq_entry.tag;
-	if (OFI_UNLIKELY(efa_rdm_write_error_msg(ep, txe->addr, prov_errno,
+	if (OFI_UNLIKELY(efa_rdm_write_error_msg(ep, txe->peer, prov_errno,
 	                                         err_msg, &err_entry.err_data_size))) {
 		err_entry.err_data_size = 0;
 	} else {
@@ -800,10 +818,10 @@ void efa_rdm_rxe_report_completion(struct efa_rdm_ope *rxe)
 	if (OFI_UNLIKELY(rxe->cq_entry.len < rxe->total_len)) {
 		EFA_WARN(FI_LOG_CQ,
 			 "Message truncated! from peer %" PRIu64
-			 " rx_id: %" PRIu32 " msg_id: %" PRIu32 " tag: %" PRIu64
+			 " implicit fi_addr: %" PRIu64 " rx_id: %" PRIu32 " msg_id: %" PRIu32 " tag: %" PRIu64
 			 " incoming message size: %" PRIu64
 			 " receiving buffer size: %zu\n",
-			 rxe->addr, rxe->rx_id, rxe->msg_id, rxe->cq_entry.tag,
+			 rxe->peer->conn->fi_addr, rxe->peer->conn->implicit_fi_addr, rxe->rx_id, rxe->msg_id, rxe->cq_entry.tag,
 			 rxe->total_len, rxe->cq_entry.len);
 
 		ret = ofi_cq_write_error_trunc(ep->base_ep.util_ep.rx_cq,
@@ -832,15 +850,17 @@ void efa_rdm_rxe_report_completion(struct efa_rdm_ope *rxe)
 	    (ofi_need_completion(cq_flags, rxe->fi_flags) ||
 	     (rxe->cq_entry.flags & FI_MULTI_RECV))) {
 		EFA_DBG(FI_LOG_CQ,
-		       "Writing recv completion for rxe from peer: %"
-		       PRIu64 " rx_id: %" PRIu32 " msg_id: %" PRIu32
-		       " tag: %lx total_len: %" PRIu64 "\n",
-		       rxe->addr, rxe->rx_id, rxe->msg_id,
-		       rxe->cq_entry.tag, rxe->total_len);
+			"Writing recv completion for rxe from peer: %" PRIu64
+			" implicit fi_addr: %" PRIu64 " rx_id: %" PRIu32
+			" msg_id: %" PRIu32 " tag: %lx total_len: %" PRIu64
+			"\n",
+			rxe->peer->conn->fi_addr,
+			rxe->peer->conn->implicit_fi_addr, rxe->rx_id,
+			rxe->msg_id, rxe->cq_entry.tag, rxe->total_len);
 
 		efa_rdm_tracepoint(recv_end,
 			    rxe->msg_id, (size_t) rxe->cq_entry.op_context,
-			    rxe->total_len, rxe->cq_entry.tag, rxe->addr);
+			    rxe->total_len, rxe->cq_entry.tag, rxe->peer->conn->fi_addr);
 
 
 		if (ep->base_ep.util_ep.caps & FI_SOURCE)
@@ -851,7 +871,7 @@ void efa_rdm_rxe_report_completion(struct efa_rdm_ope *rxe)
 					       rxe->cq_entry.buf,
 					       rxe->cq_entry.data,
 					       rxe->cq_entry.tag,
-					       rxe->addr);
+					       rxe->peer->conn->fi_addr);
 		else
 			ret = ofi_cq_write(rx_cq,
 					   rxe->cq_entry.op_context,
@@ -935,13 +955,13 @@ void efa_rdm_txe_report_completion(struct efa_rdm_ope *txe)
 		       "Writing send completion for txe to peer: %" PRIu64
 		       " tx_id: %" PRIu32 " msg_id: %" PRIu32 " tag: %lx len: %"
 		       PRIu64 "\n",
-		       txe->addr, txe->tx_id, txe->msg_id,
+		       txe->peer->conn->fi_addr, txe->tx_id, txe->msg_id,
 		       txe->cq_entry.tag, txe->total_len);
 
 
 	efa_rdm_tracepoint(send_end,
 		    txe->msg_id, (size_t) txe->cq_entry.op_context,
-		    txe->total_len, txe->cq_entry.tag, txe->addr);
+		    txe->total_len, txe->cq_entry.tag, txe->peer->conn->fi_addr);
 
 		/* TX completions should not send peer address to util_cq */
 		if (txe->ep->base_ep.util_ep.caps & FI_SOURCE)
@@ -1347,7 +1367,7 @@ int efa_rdm_ope_post_read(struct efa_rdm_ope *ope)
 		if (OFI_UNLIKELY(!pkt_entry))
 			return -FI_EAGAIN;
 
-		efa_rdm_pke_init_read_context(pkt_entry, ope, ope->addr, ofi_buf_index(ope), 0);
+		efa_rdm_pke_init_read_context(pkt_entry, ope, ofi_buf_index(ope), 0);
 		err = efa_rdm_pke_read(pkt_entry,
 					 ope->iov[0].iov_base,
 					 0,
@@ -1363,7 +1383,7 @@ int efa_rdm_ope_post_read(struct efa_rdm_ope *ope)
 
 	if (ope->type == EFA_RDM_TXE &&
 	    ope->op == ofi_op_read_req &&
-	    ope->addr == FI_ADDR_NOTAVAIL) {
+	    ope->peer == NULL) {
 		err = efa_rdm_txe_prepare_local_read_pkt_entry(ope);
 		if (err)
 			return err;
@@ -1417,7 +1437,7 @@ int efa_rdm_ope_post_read(struct efa_rdm_ope *ope)
 				    ope->rma_iov[rma_iov_idx].len - rma_iov_offset);
 		read_once_len = MIN(read_once_len, max_read_once_len);
 
-		efa_rdm_pke_init_read_context(pkt_entry, ope, ope->addr, ofi_buf_index(ope), read_once_len);
+		efa_rdm_pke_init_read_context(pkt_entry, ope, ofi_buf_index(ope), read_once_len);
 		err = efa_rdm_pke_read(pkt_entry,
 					 (char *)ope->iov[iov_idx].iov_base + iov_offset,
 					 read_once_len,
@@ -1700,10 +1720,7 @@ int efa_rdm_rxe_post_local_read_or_queue(struct efa_rdm_ope *rxe,
 	msg_rma.rma_iov = &rma_iov;
 	msg_rma.rma_iov_count = 1;
 
-	txe = efa_rdm_rma_alloc_txe(rxe->ep,
-				    efa_rdm_ep_get_peer(rxe->ep, msg_rma.addr),
-				    &msg_rma,
-				    ofi_op_read_req,
+	txe = efa_rdm_rma_alloc_txe(rxe->ep, NULL, &msg_rma, ofi_op_read_req,
 				    0 /* flags*/);
 	if (!txe) {
 		return -FI_ENOBUFS;
@@ -1913,3 +1930,48 @@ ssize_t efa_rdm_ope_repost_ope_queued_before_handshake(struct efa_rdm_ope *ope)
 		return -FI_EINVAL;
 	}
 }
+
+int efa_rdm_ope_process_queued_ope(struct efa_rdm_ope *ope, uint16_t flag)
+{
+	int ret = 0;
+
+	assert(flag & EFA_RDM_OPE_QUEUED_FLAGS);
+
+	if (!(ope->internal_flags & flag))
+		return 0;
+
+	switch (flag) {
+	case EFA_RDM_OPE_QUEUED_BEFORE_HANDSHAKE:
+		ret = efa_rdm_ope_repost_ope_queued_before_handshake(ope);
+		--ope->ep->ope_queued_before_handshake_cnt;
+		break;
+	case EFA_RDM_OPE_QUEUED_RNR:
+		assert(!dlist_empty(&ope->queued_pkts));
+		ret = efa_rdm_ep_post_queued_pkts(ope->ep, &ope->queued_pkts);
+		break;
+	case EFA_RDM_OPE_QUEUED_CTRL:
+		ret = efa_rdm_ope_post_send(ope, ope->queued_ctrl_type);
+		break;
+	case EFA_RDM_OPE_QUEUED_READ:
+		ret = efa_rdm_ope_post_read(ope);
+		break;
+	default:
+		break;
+	}
+
+	if (OFI_UNLIKELY(ret)) {
+		if (ret == -FI_EAGAIN)
+			return ret;
+
+		assert(ope->type == EFA_RDM_TXE || ope->type == EFA_RDM_RXE);
+		if (ope->type == EFA_RDM_TXE)
+			efa_rdm_txe_handle_error(ope, -ret, FI_EFA_ERR_PKT_POST);
+		else
+			efa_rdm_rxe_handle_error(ope, -ret, FI_EFA_ERR_PKT_POST);
+		return ret;
+	}
+
+	ope->internal_flags &= ~flag;
+	dlist_remove(&ope->queued_entry);
+	return ret;
+}
