@@ -55,7 +55,7 @@ int efa_rdm_pke_fill_data(struct efa_rdm_pke *pkt_entry,
 	if (efa_both_support_zero_hdr_data_transfer(pkt_entry->ep, ope->peer)) {
 		/* zero hdr transfer only happens for eager msg (non-tagged) pkt */
 		assert(pkt_type == EFA_RDM_EAGER_MSGRTM_PKT);
-		pkt_entry->flags |= EFA_RDM_PKE_SEND_TO_USER_RECV_QP;
+		pkt_entry->flags |= EFA_RDM_PKE_SEND_TO_USER_RECV_QP | EFA_RDM_PKE_HAS_NO_BASE_HDR;
 	}
 
 	/* Only 3 categories of packets has data_size and data_offset:
@@ -396,12 +396,10 @@ void efa_rdm_pke_handle_data_copied(struct efa_rdm_pke *pkt_entry)
  *
  * @param[in]	pkt_entry	pkt entry
  * @param[in]	prov_errno	provider specific error code
- * @param[in]	peer		efa_rdm_peer struct for the receiver
  */
-void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno, struct efa_rdm_peer *peer)
+void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno)
 {
 	struct efa_rdm_ope *txe;
-	struct efa_rdm_ope *rxe;
 	struct efa_rdm_ep *ep;
 
 	int err = to_fi_errno(prov_errno);
@@ -412,9 +410,9 @@ void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno,
 	        efa_strerror(prov_errno), prov_errno);
 
 	ep = pkt_entry->ep;
-	efa_rdm_ep_record_tx_op_completed(ep, pkt_entry, peer);
+	efa_rdm_ep_record_tx_op_completed(ep, pkt_entry);
 
-	if (!peer) {
+	if (!pkt_entry->peer) {
 		/*
 		 * If peer is NULL, it means the peer has been removed from AV.
 		 * In this case, ignore this error completion.
@@ -427,41 +425,51 @@ void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno,
 	switch (pkt_entry->ope->type) {
 	case EFA_RDM_TXE:
 		txe = pkt_entry->ope;
-		if (!(pkt_entry->flags & EFA_RDM_PKE_SEND_TO_USER_RECV_QP) && efa_rdm_pke_get_base_hdr(pkt_entry)->type == EFA_RDM_HANDSHAKE_PKT) {
-			if (prov_errno == EFA_IO_COMP_STATUS_REMOTE_ERROR_RNR) {
-				/*
-				 * handshake should always be queued for RNR
-				 */
-				assert(!(peer->flags & EFA_RDM_PEER_HANDSHAKE_QUEUED));
-				peer->flags |= EFA_RDM_PEER_HANDSHAKE_QUEUED;
-				dlist_insert_tail(&peer->handshake_queued_entry,
+		if (efa_rdm_pkt_type_of(pkt_entry) == EFA_RDM_HANDSHAKE_PKT) {
+			switch (prov_errno) {
+				case EFA_IO_COMP_STATUS_REMOTE_ERROR_RNR:
+					/**
+					 * handshake should always be queued for RNR
+					 */
+					assert(!(pkt_entry->peer->flags & EFA_RDM_PEER_HANDSHAKE_QUEUED));
+					pkt_entry->peer->flags |= EFA_RDM_PEER_HANDSHAKE_QUEUED;
+					dlist_insert_tail(&pkt_entry->peer->handshake_queued_entry,
 						  &efa_rdm_ep_domain(ep)->handshake_queued_peer_list);
-			} else if (prov_errno != EFA_IO_COMP_STATUS_REMOTE_ERROR_BAD_DEST_QPN) {
-				/*
-				 * If prov_errno is EFA_IO_COMP_STATUS_REMOTE_ERROR_BAD_DEST_QPN
-				 * the peer has been destroyed. Which is normal, as peer does not
-				 * always need a handshake packet to perform its duty. (For example,
-				 * if a peer just want to sent 1 message to the ep, it does not need
-				 * handshake.) In this case, it is safe to ignore this error
-				 * completion. In all other cases, we write an eq entry because
-				 * there is no application operation associated with handshake.
-				 */
-				char ep_addr_str[OFI_ADDRSTRLEN], peer_addr_str[OFI_ADDRSTRLEN];
-				size_t buflen=0;
+					break;
+				case EFA_IO_COMP_STATUS_REMOTE_ERROR_BAD_DEST_QPN: /* fall through */
+				case FI_EFA_ERR_ESTABLISHED_RECV_UNRESP: /* fall through */
+				case FI_EFA_ERR_UNESTABLISHED_RECV_UNRESP: /* fall through */
+				case EFA_IO_COMP_STATUS_REMOTE_ERROR_ABORT: /* fall through */
+				case EFA_IO_COMP_STATUS_LOCAL_ERROR_UNRESP_REMOTE: /* fall through */
+				case EFA_IO_COMP_STATUS_LOCAL_ERROR_UNREACH_REMOTE:
+					/**
+					 * This means the peer has been destroyed. Which is normal, as peer does not
+					 * always need a handshake packet to perform its duty. (For example,
+					 * if a peer just want to sent 1 message to the ep, it does not need
+					 * handshake.) In this case, it is safe to ignore this error
+					 * completion. In all other cases, we write an eq entry because
+					 * there is no application operation associated with handshake.
+					 */
+					break;
+				default:
+				{
+					char ep_addr_str[OFI_ADDRSTRLEN], peer_addr_str[OFI_ADDRSTRLEN];
+					size_t buflen=0;
 
-				memset(&ep_addr_str, 0, sizeof(ep_addr_str));
-				memset(&peer_addr_str, 0, sizeof(peer_addr_str));
-				buflen = sizeof(ep_addr_str);
-				efa_base_ep_raw_addr_str(&ep->base_ep, ep_addr_str, &buflen);
-				buflen = sizeof(peer_addr_str);
-				efa_base_ep_get_peer_raw_addr_str(&ep->base_ep, pkt_entry->addr, peer_addr_str, &buflen);
-				EFA_WARN(FI_LOG_CQ,
-					"While sending a handshake packet, an error occurred."
-					"  Our address: %s, peer address: %s\n",
+					memset(&ep_addr_str, 0, sizeof(ep_addr_str));
+					memset(&peer_addr_str, 0, sizeof(peer_addr_str));
+					buflen = sizeof(ep_addr_str);
+					efa_base_ep_raw_addr_str(&ep->base_ep, ep_addr_str, &buflen);
+					buflen = sizeof(peer_addr_str);
+					efa_base_ep_get_peer_raw_addr_str(&ep->base_ep, pkt_entry->peer->conn->fi_addr, peer_addr_str, &buflen);
+					EFA_WARN(FI_LOG_CQ,
+						"While sending a handshake packet, an error occurred."
+						"  Our address: %s, peer address: %s\n",
 					ep_addr_str, peer_addr_str);
-				efa_base_ep_write_eq_error(&ep->base_ep, err, prov_errno);
+					efa_base_ep_write_eq_error(&ep->base_ep, err, prov_errno);
+					break;
+				}
 			}
-
 			efa_rdm_pke_release_tx(pkt_entry);
 			efa_rdm_txe_release(txe);
 
@@ -492,12 +500,7 @@ void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno,
 				 * packets include all REQ, DATA) thus shoud be queued for RNR
 				 * only if application wants EFA to manager resource.
 				 */
-				efa_rdm_ep_queue_rnr_pkt(ep, &txe->queued_pkts, pkt_entry, peer);
-				if (!(txe->internal_flags & EFA_RDM_OPE_QUEUED_RNR)) {
-					txe->internal_flags |= EFA_RDM_OPE_QUEUED_RNR;
-					dlist_insert_tail(&txe->queued_entry,
-							  &efa_rdm_ep_domain(ep)->ope_queued_list);
-				}
+				efa_rdm_ep_queue_rnr_pkt(ep, pkt_entry);
 			}
 		} else {
 			efa_rdm_txe_handle_error(pkt_entry->ope, err, prov_errno);
@@ -505,7 +508,6 @@ void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno,
 		}
 		break;
 	case EFA_RDM_RXE:
-		rxe = pkt_entry->ope;
 		if (prov_errno == EFA_IO_COMP_STATUS_REMOTE_ERROR_RNR) {
 			/*
 			 * This packet is associated with a recv operation, (such packets
@@ -513,12 +515,7 @@ void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno,
 			 * is regardless value of ep->handle_resource_management, because
 			 * resource management is only applied to send operation.
 			 */
-			efa_rdm_ep_queue_rnr_pkt(ep, &rxe->queued_pkts, pkt_entry, peer);
-			if (!(rxe->internal_flags & EFA_RDM_OPE_QUEUED_RNR)) {
-				rxe->internal_flags |= EFA_RDM_OPE_QUEUED_RNR;
-				dlist_insert_tail(&rxe->queued_entry,
-						  &efa_rdm_ep_domain(ep)->ope_queued_list);
-			}
+			efa_rdm_ep_queue_rnr_pkt(ep, pkt_entry);
 		} else {
 			efa_rdm_rxe_handle_error(pkt_entry->ope, err, prov_errno);
 			efa_rdm_pke_release_tx(pkt_entry);
@@ -543,24 +540,23 @@ void efa_rdm_pke_handle_tx_error(struct efa_rdm_pke *pkt_entry, int prov_errno,
  * then release the packet entry.
  *
  * @param[in,out]	pkt_entry	packet entry
- * @param[in,out]	peer	efa_rdm_peer struct of the receiver
  */
-void efa_rdm_pke_handle_send_completion(struct efa_rdm_pke *pkt_entry, struct efa_rdm_peer *peer)
+void efa_rdm_pke_handle_send_completion(struct efa_rdm_pke *pkt_entry)
 {
 	struct efa_rdm_ep *ep;
 
 	ep = pkt_entry->ep;
 	/*
-	 * For a send completion, pkt_entry->addr can be FI_ADDR_NOTAVAIL in 3 situations:
+	 * For a send completion, pkt_entry->peer can be NULL in 3 situations:
 	 * 1. the pkt_entry is used for a local read operation
 	 * 2. a new peer with same gid+qpn was inserted to av, thus the peer was removed from AV.
 	 * 3. application removed the peer's address from av.
 	 * In 1, we should proceed. For 2 and 3, the send completion should be ignored.
 	 */
-	if (pkt_entry->addr == FI_ADDR_NOTAVAIL &&
+	if (!pkt_entry->peer &&
 	    !(pkt_entry->flags & EFA_RDM_PKE_LOCAL_READ)) {
 		EFA_WARN(FI_LOG_CQ, "ignoring send completion of a packet to a removed peer.\n");
-		efa_rdm_ep_record_tx_op_completed(ep, pkt_entry, peer);
+		efa_rdm_ep_record_tx_op_completed(ep, pkt_entry);
 		efa_rdm_pke_release_tx(pkt_entry);
 		return;
 	}
@@ -568,13 +564,13 @@ void efa_rdm_pke_handle_send_completion(struct efa_rdm_pke *pkt_entry, struct ef
 	/* These pkts are eager pkts withour hdrs */
 	if (pkt_entry->flags & EFA_RDM_PKE_SEND_TO_USER_RECV_QP) {
 		efa_rdm_pke_handle_eager_rtm_send_completion(pkt_entry);
-		efa_rdm_ep_record_tx_op_completed(ep, pkt_entry, peer);
+		efa_rdm_ep_record_tx_op_completed(ep, pkt_entry);
 		efa_rdm_pke_release_tx(pkt_entry);
 		return;
 	}
 
 	/* Start handling pkts with hdrs */
-	switch (efa_rdm_pke_get_base_hdr(pkt_entry)->type) {
+	switch (efa_rdm_pkt_type_of(pkt_entry)) {
 	case EFA_RDM_HANDSHAKE_PKT:
 		efa_rdm_txe_release(pkt_entry->ope);
 		break;
@@ -590,7 +586,7 @@ void efa_rdm_pke_handle_send_completion(struct efa_rdm_pke *pkt_entry, struct ef
 		efa_rdm_pke_handle_eor_send_completion(pkt_entry);
 		break;
 	case EFA_RDM_RMA_CONTEXT_PKT:
-		efa_rdm_pke_handle_rma_completion(pkt_entry, peer);
+		efa_rdm_pke_handle_rma_completion(pkt_entry);
 		return;
 	case EFA_RDM_ATOMRSP_PKT:
 		efa_rdm_pke_handle_atomrsp_send_completion(pkt_entry);
@@ -674,7 +670,7 @@ void efa_rdm_pke_handle_send_completion(struct efa_rdm_pke *pkt_entry, struct ef
 		return;
 	}
 
-	efa_rdm_ep_record_tx_op_completed(ep, pkt_entry, peer);
+	efa_rdm_ep_record_tx_op_completed(ep, pkt_entry);
 	efa_rdm_pke_release_tx(pkt_entry);
 }
 
@@ -736,57 +732,11 @@ void efa_rdm_pke_handle_rx_error(struct efa_rdm_pke *pkt_entry, int prov_errno)
 	efa_rdm_pke_release_rx(pkt_entry);
 }
 
-static
-fi_addr_t efa_rdm_pke_insert_addr(struct efa_rdm_pke *pkt_entry, void *raw_addr)
-{
-	int ret;
-	fi_addr_t rdm_addr;
-	struct efa_rdm_ep *ep;
-	struct efa_rdm_base_hdr *base_hdr;
-
-	ep = pkt_entry->ep;
-
-	base_hdr = efa_rdm_pke_get_base_hdr(pkt_entry);
-	if (base_hdr->version < EFA_RDM_PROTOCOL_VERSION) {
-		char self_raw_addr_str[OFI_ADDRSTRLEN];
-		size_t buflen = OFI_ADDRSTRLEN;
-
-		efa_base_ep_raw_addr_str(&ep->base_ep, self_raw_addr_str, &buflen);
-		EFA_WARN(FI_LOG_CQ,
-			"Host %s received a packet with invalid protocol version %d.\n"
-			"This host can only support protocol version %d and above.\n",
-			self_raw_addr_str, base_hdr->version, EFA_RDM_PROTOCOL_VERSION);
-		efa_base_ep_write_eq_error(&ep->base_ep, FI_EIO, FI_EFA_ERR_INVALID_PKT_TYPE);
-		fprintf(stderr, "Host %s received a packet with invalid protocol version %d.\n"
-			"This host can only support protocol version %d and above. %s:%d\n",
-			self_raw_addr_str, base_hdr->version, EFA_RDM_PROTOCOL_VERSION, __FILE__, __LINE__);
-		abort();
-	}
-
-	assert(base_hdr->type >= EFA_RDM_REQ_PKT_BEGIN);
-
-	/*
-	 * The message is from a peer through efa device, which means peer is not local
-	 * or shm is disabled for transmission.
-	 * We shouldn't insert shm av anyway in this case.
-	 * Also, calling fi_av_insert internally inside progress engine is violating
-	 * Libfabric standard for FI_AV_TABLE.
-	 */
-	ret = efa_av_insert_one(ep->base_ep.av, (struct efa_ep_addr *)raw_addr,
-	                        &rdm_addr, 0, NULL, false);
-	if (OFI_UNLIKELY(ret != 0)) {
-		efa_base_ep_write_eq_error(&ep->base_ep, ret, FI_EFA_ERR_AV_INSERT);
-		return -1;
-	}
-
-	return rdm_addr;
-}
-
 void efa_rdm_pke_proc_received_no_hdr(struct efa_rdm_pke *pkt_entry, bool has_imm_data, uint32_t imm_data)
 {
 	struct efa_rdm_ope *rxe = pkt_entry->ope;
 
-	assert(pkt_entry->alloc_type == EFA_RDM_PKE_FROM_USER_RX_POOL);
+	assert(pkt_entry->flags & EFA_RDM_PKE_HAS_NO_BASE_HDR);
 	assert(rxe);
 
 	if (has_imm_data) {
@@ -794,7 +744,8 @@ void efa_rdm_pke_proc_received_no_hdr(struct efa_rdm_pke *pkt_entry, bool has_im
 		rxe->cq_entry.data = imm_data;
 	}
 
-	rxe->addr = pkt_entry->addr;
+	rxe->peer = pkt_entry->peer;
+	dlist_insert_tail(&rxe->peer_entry, &rxe->peer->rxe_list);
 	rxe->total_len = pkt_entry->pkt_size;
 	rxe->cq_entry.len = pkt_entry->pkt_size;
 
@@ -810,7 +761,7 @@ void efa_rdm_pke_proc_received_no_hdr(struct efa_rdm_pke *pkt_entry, bool has_im
  * @param[in]	pkt_entry	received packet entry
  * @param[in]	peer		peer struct of the sender
  */
-void efa_rdm_pke_proc_received(struct efa_rdm_pke *pkt_entry, struct efa_rdm_peer *peer)
+void efa_rdm_pke_proc_received(struct efa_rdm_pke *pkt_entry)
 {
 	struct efa_rdm_ep *ep;
 	struct efa_rdm_base_hdr *base_hdr;
@@ -844,7 +795,7 @@ void efa_rdm_pke_proc_received(struct efa_rdm_pke *pkt_entry, struct efa_rdm_pee
 		efa_rdm_pke_handle_eor_recv(pkt_entry);
 		return;
 	case EFA_RDM_HANDSHAKE_PKT:
-		efa_rdm_pke_handle_handshake_recv(pkt_entry, peer);
+		efa_rdm_pke_handle_handshake_recv(pkt_entry);
 		return;
 	case EFA_RDM_CTS_PKT:
 		efa_rdm_pke_handle_cts_recv(pkt_entry);
@@ -881,7 +832,7 @@ void efa_rdm_pke_proc_received(struct efa_rdm_pke *pkt_entry, struct efa_rdm_pee
 	case EFA_RDM_DC_WRITE_RTA_PKT:
 	case EFA_RDM_FETCH_RTA_PKT:
 	case EFA_RDM_COMPARE_RTA_PKT:
-		efa_rdm_pke_handle_rtm_rta_recv(pkt_entry, peer);
+		efa_rdm_pke_handle_rtm_rta_recv(pkt_entry);
 		return;
 	case EFA_RDM_EAGER_RTW_PKT:
 		efa_rdm_pke_handle_eager_rtw_recv(pkt_entry);
@@ -913,24 +864,3 @@ void efa_rdm_pke_proc_received(struct efa_rdm_pke *pkt_entry, struct efa_rdm_pee
 		return;
 	}
 }
-
-/**
- * @brief Read peer raw address from packet header and insert the peer in AV.
- * @param ep Pointer to RDM endpoint
- * @param pkt_entry Pointer to packet entry
- * @returns Peer address, or FI_ADDR_NOTAVIL if the packet header does not include raw address
- */
-fi_addr_t efa_rdm_pke_determine_addr(struct efa_rdm_pke *pkt_entry)
-{
-	struct efa_rdm_base_hdr *base_hdr;
-
-	base_hdr = efa_rdm_pke_get_base_hdr(pkt_entry);
-	if (base_hdr->type >= EFA_RDM_REQ_PKT_BEGIN && efa_rdm_pke_get_req_raw_addr(pkt_entry)) {
-		void *raw_addr;
-		raw_addr = efa_rdm_pke_get_req_raw_addr(pkt_entry);
-		assert(raw_addr);
-		return efa_rdm_pke_insert_addr(pkt_entry, raw_addr);
-	}
-
-	return FI_ADDR_NOTAVAIL;
-}
