@@ -319,6 +319,7 @@ int vrb_set_rai(uint32_t addr_format, void *src_addr, size_t src_addrlen,
 		rai->ai_flags |= RAI_FAMILY;
 		break;
 	case FI_SOCKADDR:
+	case FI_SOCKADDR_IP:
 		rai->ai_port_space = RDMA_PS_TCP;
 		if (src_addr && src_addrlen) {
 			rai->ai_family = ((struct sockaddr *)src_addr)->sa_family;
@@ -532,6 +533,56 @@ static const char *vrb_link_layer_str(uint8_t link_layer)
 	}
 }
 
+static void vrb_get_device_bus_attr(const char *dev_name, struct fi_bus_attr *bus_attr)
+{
+	char *path = NULL;
+	char *real_path;
+	char *pci;
+	int n;
+	uint32_t a = 0, b = 0, c = 0, d = 0;
+
+	if (asprintf(&path, "/sys/class/infiniband/%s/device", dev_name) < 0 ||
+	    !path) {
+		VRB_WARN(FI_LOG_FABRIC,
+			 "Unable to allocate memory for PCI device path\n");
+		return;
+	}
+
+	real_path = realpath(path, NULL);
+	if (!real_path) {
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "realpath");
+		goto free_path;
+	}
+
+	pci = strrchr(real_path, '/');
+	if (!pci) {
+		VRB_WARN(FI_LOG_FABRIC,
+			 "Unable to find PCI device for %s\n", dev_name);
+		goto free_real_path;
+	}
+
+	pci++; /* skip '/' */
+
+	n = sscanf(pci, "%x:%x:%x.%x", &a, &b, &c, &d);
+	if (n != 4) {
+		VRB_WARN(FI_LOG_FABRIC,
+			 "Failed to parse pci bus address: %s\n", pci);
+		goto free_real_path;
+	}
+
+	bus_attr->attr.pci.domain_id = a;
+	bus_attr->attr.pci.bus_id = b;
+	bus_attr->attr.pci.device_id = c;
+	bus_attr->attr.pci.function_id = d;
+	bus_attr->bus_type = FI_BUS_PCI;
+
+free_real_path:
+	free(real_path);
+
+free_path:
+	free(path);
+}
+
 static int vrb_get_device_attrs(struct ibv_context *ctx,
 				   struct fi_info *info, uint32_t protocol)
 {
@@ -676,6 +727,8 @@ static int vrb_get_device_attrs(struct ibv_context *ctx,
 		return -FI_ENOMEM;
 	}
 
+	vrb_get_device_bus_attr(dev_name, info->nic->bus_attr);
+
 	return 0;
 }
 
@@ -727,10 +780,12 @@ static bool vrb_hmem_supported(const char *dev_name)
 	if (ofi_hmem_p2p_disabled())
 		return false;
 
-	if (vrb_gl_data.peer_mem_support && strstr(dev_name, "mlx"))
+	if (vrb_gl_data.peer_mem_support && (strstr(dev_name, "mlx") ||
+					     strstr(dev_name, "bnxt_re")))
 		return true;
 
-	if (vrb_gl_data.dmabuf_support && strstr(dev_name, "mlx5"))
+	if (vrb_gl_data.dmabuf_support && (strstr(dev_name, "mlx") ||
+					   strstr(dev_name, "bnxt_re")))
 		return true;
 
 	return false;
@@ -765,7 +820,7 @@ static int vrb_alloc_info(struct ibv_context *ctx, struct fi_info **info,
 		fi->caps = VERBS_MSG_CAPS;
 		*(fi->tx_attr) = verbs_tx_attr;
 		*(fi->rx_attr) = verbs_rx_attr;
-		fi->addr_format = FI_SOCKADDR_IB;
+		fi->addr_format = FI_SOCKADDR;
 		break;
 	case FI_EP_DGRAM:
 		fi->caps = VERBS_DGRAM_CAPS;
